import numpy as np
from numpy.linalg import norm
from scipy.spatial.transform import Rotation
from typing import TypeVar, Generic, Sequence, Callable


# proper inserstions by key with bicest module in python <3.10
# https://stackoverflow.com/questions/27672494/how-to-use-bisect-insort-left-with-a-key

T = TypeVar('T')
V = TypeVar('V')

class KeyWrapper(Generic[T, V]):
    def __init__(self, iterable: Sequence[T], key: Callable[[T], V]):
        self.it = iterable
        self.key = key

    def __getitem__(self, i: int) -> V:
        return self.key(self.it[i])

    def __len__(self) -> int:
        return len(self.it)

def vector_angle(v1, v2):
    """Find an angle between two 2D vectors"""
    v1, v2 = np.asarray(v1), np.asarray(v2)
    cos = np.dot(v1, v2) / (norm(v1) * norm(v2))
    angle = np.arccos(cos) 
    # Cross to indicate correct relative orienataion of v2 w.r.t. v1
    cross = np.cross(v1, v2)
    
    if abs(cross) > 1e-5:
        angle *= np.sign(cross)
    return angle

def R2D(angle):
    """2D rotation matrix by an angle"""
    return np.array([[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]])

def vector_align_3D(v1, v2):
    """Find a rotation to align v1 with v2"""

    v1, v2 = np.asarray(v1), np.asarray(v2)
    cos = np.dot(v1, v2) / (norm(v1) * norm(v2))
    cos = max(min(cos, 1), -1)  # NOTE: getting rid of numbers like 1.000002 that appear due to numerical instability

    angle = np.arccos(cos) 

    # Cross to get the axis of rotation
    cross = np.cross(v1, v2)
    cross = cross / norm(cross)

    return Rotation.from_rotvec(cross * angle)


def close_enough(f1, f2=0, tol=1e-4):
    """Compare two floats correctly """
    return abs(f1 - f2) < tol

def bbox_paths(paths):
    """Bounding box of a list of paths/Edge Sequences"""

    bboxes = np.array([p.bbox() for p in paths])
    return (min(bboxes[:, 0]), max(bboxes[:, 1]), min(bboxes[:, 2]), max(bboxes[:, 3]))

def lin_interpolation(val1, val2, factor):
    """Linear interpolation between val1 and val2 with factor [0, 1]

    with factor == 0, output is val1
    with factor == 1, output is val2
    """
    if factor < 0 or factor > 1:
        raise ValueError(f'lin_interpolation::Error::Expected a factor \in [0, 1], got {factor}')

    return (1 - factor) * val1 + factor * val2


# ---- Complex numbers converters ----- 
def c_to_list(num):
    """Convert complex number to a list of 2 elements
        Allows processing of lists of complex numbers
    """

    if isinstance(num, (list, tuple, set, np.ndarray)):
        return [c_to_list(n) for n in num]
    else: 
        return [num.real, num.imag]

def c_to_np(num):
    """Convert complex number to a numpy array of 2 elements
        Allows processing of lists of complex numbers
    """
    if isinstance(num, (list, tuple, set, np.ndarray)):
        return np.asarray([c_to_list(n) for n in num])
    else: 
        return np.asarray([num.real, num.imag])

def list_to_c(num):
    """Convert 2D list or list of 2D lists into complex number/list of complex numbers"""
    if isinstance(num[0], (list, tuple, set, np.ndarray)):
        return [complex(n[0], n[1]) for n in num]
    else: 
        return complex(num[0], num[1])
    
# ---- Nested Dictionaries shortcuts ---- 
# https://stackoverflow.com/a/37704379
def nested_get(dic, keys):    
    for key in keys:
        dic = dic[key]
    return dic

def nested_set(dic, keys, value):
    for key in keys[:-1]:
        dic = dic.setdefault(key, {})
    dic[keys[-1]] = value

def nested_del(dic, keys):
    for key in keys[:-1]:
        dic = dic[key]
    del dic[keys[-1]]



"""Shortcuts for common operations on panels and components"""

from copy import deepcopy, copy
import numpy as np
from numpy.linalg import norm
from scipy.spatial.transform import Rotation as R
from scipy.optimize import minimize
import matplotlib.pyplot as plt
import svgpathtools as svgpath

# Custom 
from .edge import Edge, CurveEdge, CircleEdge, EdgeSequence
from .interface import Interface
from .generic_utils import vector_angle, close_enough, c_to_list, c_to_np, list_to_c
from .base import BaseComponent
from . import flags

# ANCHOR ----- Edge Sequences Modifiers ----
def cut_corner(target_shape:EdgeSequence, target_interface:Interface):
    """ Cut the corner made of edges 1 and 2 following the shape of target_shape
        This routine updated the panel geometry and interfaces appropriately

        Parameters:
        * 'target_shape' is an EdgeSequence that is expected to contain one Edge or sequence of chained Edges 
            (next one starts from the end vertex of the one before)
            # NOTE: 'target_shape' might be scaled (along the main direction) to fit the corner size
        * Panel to modify
        * target_interface -- the chained pairs of edges that form the corner to cut, s.t. the end vertex of eid1 is at the corner
            # NOTE: Onto edges are expected to be straight lines for simplicity

        # NOTE There might be slight computational errors in the resulting shape, 
                that are more pronounced on svg visualizations due to scaling and rasterization

        Side-Effects:
            * Modified the panel shape to insert new edges
            * Adds new interface object corresponding to new edges to the panel interface list

        Returns:
            * Newly inserted edges
            * New interface object corresponding to new edges
    """
    # ---- Evaluate optimal projection of the target shape onto the corner
    corner_shape = target_shape.copy()
    panel = target_interface.panel[0]   
    target_edges = target_interface.edges
    
    # Get rid of directions by working on vertices
    if target_edges[0].start is target_edges[-1].end:
        # Orginal edges have beed reversed in normalization or smth
        target_edges.edges.reverse()  # UPD the order

    if corner_shape[0].start is corner_shape[-1].end:
        # Orginal edges have beed reversed in normalization or smth
        corner_shape.edges.reverse()  # UPD the order

    if corner_shape[0].start[1] > corner_shape[-1].end[1]:
        # now corner shape is oriented the same way as vertices
        corner_shape.reverse()
        corner_shape.snap_to([0,0])

    shortcut = corner_shape.shortcut()

    # Curves  (can be defined outside)
    curve1 = target_edges[0].as_curve()
    curve2 = target_edges[1].as_curve()

    # align order with the a projecting shape, s.t. 
    # curve2 is alawys the lower one
    swaped = False
    if target_edges[0].start[1] > target_edges[-1].end[1]:
        curve1, curve2 = curve2, curve1
        swaped = True
        # NOW curve1 is lower then curve2

    # ----- FIND OPTIMAL PLACE -----
    start = [0.5, 0.5]
    out = minimize(
       _fit_location_corner, start, 
       args=(shortcut[1] - shortcut[0], curve1, curve2),
       bounds=[(0, 1), (0, 1)])
    
    if flags.VERBOSE and not out.success:
        print(f'Cut_corner::Error::finding the projection (translation) is unsuccessful. Likely an error in edges choice')
        print(out)

    if flags.VERBOSE and not close_enough(out.fun):
        print(f'Cut_corner::Warning::projection on {target_interface} finished with fun={out.fun}')
        print(out) 

    loc = out.x
    point1 = c_to_list(curve1.point(loc[0]))
    # re-align corner_shape with found shifts
    corner_shape.snap_to(point1)   
    
    # ----- UPD panel ----
    # Complete to the full corner -- connect with the initial vertices
    if swaped:
        # The edges are aligned as v2 -> vc -> v1
        corner_shape.reverse()
        loc[0], loc[1] = loc[1], loc[0]

    # Insert a new shape
    cut_edge1, _ = target_edges[0].subdivide_param([loc[0], 1-loc[0]])
    _,  cut_edge2 = target_edges[1].subdivide_param([loc[1], 1-loc[1]])
    
    cut_edge1.end = corner_shape[0].start  # Connect with new insert
    cut_edge2.start = corner_shape[-1].end

    corner_shape.insert(0, cut_edge1)
    corner_shape.append(cut_edge2)

    # Substitute edges in the panel definition
    panel.edges.pop(target_edges[0])
    panel.edges.substitute(target_edges[1], corner_shape)

    # Update interface definitions
    target_edges = EdgeSequence(target_edges.edges)  # keep the same edge references, 
                                                     # but not the same edge sequence reference
                                                     # In case it matches one of the interfaces (we don't want target edges to be overriden)
    iter = panel.interfaces if isinstance(panel.interfaces, list) else panel.interfaces.values() 
    for intr in iter:
        # Substitute old edges with what's left from them after cutting
        if target_edges[0] in intr.edges:
            intr.edges.substitute(target_edges[0], corner_shape[0])
        if target_edges[1] in intr.edges:
            intr.edges.substitute(target_edges[1], corner_shape[-1])

    # Add new interface corresponding to the introduced cut
    new_int = Interface(panel, corner_shape[1:-1])
    if isinstance(panel.interfaces, list):
        panel.interfaces.append(new_int)
    else:
        panel.interfaces[f'int_{len(panel.interfaces)}'] = new_int   

    return corner_shape[1:-1], new_int

def cut_into_edge(target_shape, base_edge:Edge, offset=0, right=True, tol=1e-4):
    """ Insert edges of the target_shape into the given base_edge, starting from offset
        edges in target shape are rotated s.t. start -> end vertex vector is aligned with the edge 

        NOTE: for now the base_edge is treated as straight edge

        Parameters:
        * target_shape -- list of single edge or chained edges to be inserted in the edge. 
        * base_edge -- edge object, defining the border
        * right -- which direction the cut should be oriented w.r.t. the direction of base edge
        * Offset -- position of the center of the target shape along the edge.  

        Returns:
        * Newly created edges that accomodate the cut
        * Edges corresponding to the target shape
        * Edges that lie on the original base edge 
    """
    target_shape = EdgeSequence(target_shape)
    new_edges = target_shape.copy().snap_to([0, 0])  # copy and normalize translation of vertices

    # Simplify to vectors
    shortcut = new_edges.shortcut()  # "Interface" of the shape to insert
    target_shape_w = norm(shortcut)
    edge_len = base_edge.length()

    if offset < target_shape_w / 2  - tol or offset > (edge_len - target_shape_w / 2) + tol:   
        # NOTE: This is not a definitive check, and the cut might still not fit, depending on the base_edge curvature
        raise ValueError(f'Operators-CutingIntoEdge::Error::offset value is not within the base_edge length')

    # find starting vertex for insertion & place edges there
    curve = base_edge.as_curve()
    rel_offset = curve.ilength(offset)   

    # ----- OPTIMIZATION --- 
    start = [0.1, 0.1]
    out = minimize(
       _fit_location_edge, start, 
       args=(rel_offset, target_shape_w, curve),
       bounds=[(0, 1)])
    shift = out.x  

    # Error checks
    if flags.VERBOSE and not out.success:
        print(f'Cut_edge::Error::finding the projection (translation) is unsuccessful. Likely an error in edges choice')

    if not close_enough(out.fun, tol=0.01):
        if flags.VERBOSE:
            print(out) 
        raise RuntimeError(f'Cut_edge::ERROR::projection on {base_edge} finished with fun={out.fun}')
    
    if rel_offset + shift[0] > 1 + tol or (rel_offset - shift[1]) < 0 - tol:
        raise RuntimeError(
            f'Cut_edge::ERROR::projection on {base_edge} is out of edge bounds: '
            f'[{rel_offset - shift[1], rel_offset + shift[0]}].'
            ' Check the offset value')

    # All good -- integrate the target shape
    ins_point = c_to_np(curve.point(rel_offset - shift[1])) if (rel_offset - shift[1]) > tol else base_edge.start
    fin_point = c_to_np(curve.point(rel_offset + shift[0])) if (rel_offset + shift[0]) < 1 - tol else base_edge.end

    # Align the shape with an edge
    # find rotation to apply on target shape 
    insert_vector = np.asarray(fin_point) - np.asarray(ins_point)
    angle = vector_angle(shortcut[1] - shortcut[0], insert_vector)
    new_edges.rotate(angle) 

    # place
    new_edges.snap_to(ins_point)

    # Check orientation 
    avg_vertex = np.asarray(new_edges.verts()).mean(0)
    right_position = np.sign(np.cross(insert_vector, avg_vertex - np.asarray(new_edges[0].start))) == -1 
    if not right and right_position or right and not right_position:
        # flip shape to match the requested direction
        new_edges.reflect(new_edges[0].start, new_edges[-1].end)  

    # Integrate edges
    # NOTE: no need to create extra edges if the the shape is incerted right at the beggining or end of the edge
    base_edge_leftovers = EdgeSequence()
    start_id, end_id = 0, len(new_edges)

    if ins_point is base_edge.start:
        new_edges[0].start = base_edge.start   # Connect into the original edge
    else:
        start_part = base_edge.subdivide_param([rel_offset - shift[1], 1 - (rel_offset - shift[1])])[0]
        start_part.end = new_edges[0].start
        new_edges.insert(0, start_part)
        base_edge_leftovers.append(new_edges[0])
        start_id = 1 

    if fin_point is base_edge.end:
        new_edges[-1].end = base_edge.end  # Connect into the original edge
    else:
        end_part = base_edge.subdivide_param([rel_offset + shift[0], 1 - (rel_offset + shift[0])])[-1]
        end_part.start = new_edges[-1].end
        new_edges.append(end_part)
        base_edge_leftovers.append(new_edges[-1])
        end_id = -1
        
    return new_edges, new_edges[start_id:end_id], base_edge_leftovers

def _fit_location_corner(l, diff_target, curve1, curve2):
    """Find the points on two curves s.t. vector between them is the same as shortcut"""

    # Current points on curves
    point1 = c_to_np(curve1.point(l[0]))
    point2 = c_to_np(curve2.point(l[1]))
    diff_curr = point2 - point1

    if flags.VERBOSE:
        print('Location Progression: ', (diff_curr[0] - diff_target[0])**2, (diff_curr[1] - diff_target[1])**2)

    return ((diff_curr[0] - diff_target[0])**2 
            + (diff_curr[1] - diff_target[1])**2)

def _fit_location_edge(shift, location, width_target, curve):
    """Find the points on two curves s.t. vector between them is the same as shortcut"""

    # Current points on curves
    pointc = c_to_np(curve.point(location))   
    point1 = c_to_np(curve.point(location + shift[0]))
    point2 = c_to_np(curve.point(location - shift[1]))


    if flags.VERBOSE:
        print('Location Progression: ', (_dist(point1, point2) - width_target)**2)

    # regularize points to be at the same distance from center
    reg_symmetry = (_dist(point1, pointc) - _dist(point2, pointc))**2

    return (_dist(point1, point2) - width_target)**2 + reg_symmetry


# ANCHOR ----- Panel operations ------
def distribute_Y(component, n_copies, odd_copy_shift=10):
    """Distribute copies of component over the circle around Oy"""
    copies = [ component ]
    delta_rotation = R.from_euler('XYZ', [0, 360 / n_copies, 0], degrees=True)
    
    for i in range(n_copies - 1):
        new_component = deepcopy(copies[-1])
        new_component.name = f'panel_{i}'   # Unique
        new_component.rotate_by(delta_rotation)
        new_component.translate_to(delta_rotation.apply(new_component.translation))

        copies.append(new_component)

    # shift around to resolve collisions (hopefully)
    if odd_copy_shift:
        for i in range(n_copies):
            if not i % 2:
                copies[i].translate_by(copies[i].norm() * odd_copy_shift)
        
    return copies

def distribute_horisontally(component, n_copies, stride=20, name_tag='panel'):
    """Distribute copies of component over the straight horisontal line perpendicular to the norm"""
    copies = [ component ]
    component.name = f'{name_tag}_0'   # Unique

    if isinstance(component, BaseComponent):
        translation_dir = component.rotation.apply([0, 0, 1])   # Horisontally along the panel
        translation_dir = np.cross(translation_dir, [0, 1, 0])   # perpendicular to Y
        translation_dir = translation_dir / norm(translation_dir)
        delta_translation = translation_dir * stride
    else:
        translation_dir = [1, 0, 0] 

    for i in range(n_copies - 1):
        new_component = deepcopy(copies[-1])  
        new_component.name = f'{name_tag}_{i + 1}'   # Unique
        new_component.translate_by(delta_translation)

        copies.append(new_component)

    return copies


# ANCHOR ----- Sleeve support -----
def even_armhole_openings(front_opening, back_opening, tol=1e-2):
    """
        Rearrange sleeve openings for front and back s.t. their projection 
        on vertical line is the same, while preserving the overall shape.
        Allows for creation of two symmetric sleeve panels from them

        !! Important: assumes that the front opening is longer then back opening
    """
    # Construct sleeve panel shapes from opening inverses
    cfront, cback = front_opening.copy(), back_opening.copy()
    cback.reflect([0, 0], [1, 0]).reverse().snap_to(cfront[-1].end)

    # Cutout
    slope = np.array([cfront[0].start, cback[-1].end])
    slope_vec = slope[1] - slope[0]
    slope_perp = np.asarray([-slope_vec[1], slope_vec[0]])
    slope_midpoint = (slope[0] + slope[1]) / 2

    # Intersection with the sleeve itself line
    # svgpath tools allow solution regardless of egde types
    inter_segment = svgpath.Line(
        list_to_c(slope_midpoint - 20 * slope_perp), 
        list_to_c(slope_midpoint + 20 * slope_perp)
    )
    target_segment = cfront[-1].as_curve()

    intersect_t = target_segment.intersect(inter_segment)
    if len(intersect_t) != 1:
        print(
            f'Sleeve Opening Inversion::Error::{len(intersect_t)} intersection points instead of one. '
            'Check the quality of supplied curves'
        )
    
    if (len(intersect_t) >= 1 
            and not (close_enough(intersect_t[0][0], 0, tol=tol)   # checking if they are already ok separated
                     or close_enough(intersect_t[0][0], 1, tol=tol))):
        # The current separation is not satisfactory
        # Update the opening shapes
        intersect_t = intersect_t[0][0]
        subdiv = front_opening.edges[-1].subdivide_param([intersect_t, 1 - intersect_t])
        front_opening.substitute(-1, subdiv[0])  

        # Move this part to the back opening
        subdiv[1].start, subdiv[1].end = copy(subdiv[1].start), copy(subdiv[1].end)  # Disconnect vertices in subdivided version
        subdiv.pop(0) 
        subdiv.reflect([0, 0], [1, 0]).reverse().snap_to(back_opening[-1].end)
        subdiv[0].start = back_opening[-1].end
        
        back_opening.append(subdiv[0])

    # Align the slope with OY direction
    # for correct size of sleeve panels
    slope_angle = np.arctan(-slope_vec[0] / slope_vec[1])	
    front_opening.rotate(-slope_angle)
    back_opening.rotate(slope_angle)

    return front_opening, back_opening

# ANCHOR ----- Curve tools -----
def _max_curvature(curve, points_estimates=100):
    """Average curvature in a curve"""
    # NOTE: this work slow, but direct evaluation seems
    # infeasible
    # Some hints here: https://math.stackexchange.com/questions/1954845/bezier-curvature-extrema
    t_space = np.linspace(0, 1, points_estimates)
    return max([curve.curvature(t) for t in t_space])

def _bend_extend_2_tangent(
        shift, cp, target_len, direction, 
        target_tangent_start, target_tangent_end, 
        point_estimates=50):
    """Evaluate how well curve preserves the length and tangents

        NOTE: point_estimates controls average curvature evaluation.
            The higher the number, the more stable the optimization,
            but higher computational cost
    """

    control = np.array([
        cp[0], 
        [cp[1][0] + shift[0], cp[1][1] + shift[1]], 
        [cp[2][0] + shift[2], cp[2][1] + shift[3]],
        cp[-1] + direction * shift[4]
    ])

    params = control[:, 0] + 1j*control[:, 1]
    curve_inverse = svgpath.CubicBezier(*params)

    length_diff = (curve_inverse.length() - target_len)**2  # preservation

    tan_0_diff = (abs(curve_inverse.unit_tangent(0) - target_tangent_start))**2
    tan_1_diff = (abs(curve_inverse.unit_tangent(1) - target_tangent_end))**2

    # NOTE: tried regularizing based on Y value in relative coordinates (for speed), 
    # But it doesn't produce good results
    curvature_reg = _max_curvature(curve_inverse, points_estimates=point_estimates)**2

    end_expantion_reg = 0.001*shift[-1]**2 

    return length_diff + tan_0_diff + tan_1_diff + curvature_reg + end_expantion_reg
      
def curve_match_tangents(curve, target_tan0, target_tan1, return_as_edge=False):
    """Update the curve to have the desired tangent directions at endpoints 
        while preserving curve length and overall direction

        Returns 
        * control points for the final CubicBezier curves
        * Or CurveEdge instance, if return_as_edge=True

        NOTE: Only Cubic Bezier curves are supported
    """
    if not isinstance(curve, svgpath.CubicBezier):
        raise NotImplementedError(
            f'Curve_match_tangents::Error::Only Cubic Bezier curves are supported ', 
            f'(got {type(curve)})')

    curve_cps = c_to_np(curve.bpoints())

    direction = curve_cps[-1] - curve_cps[0]
    direction /= np.linalg.norm(direction)

    target_tan0 = target_tan0 / np.linalg.norm(target_tan0)
    target_tan1 = target_tan1 / np.linalg.norm(target_tan1)

    # match tangents with the requested ones while preserving length
    out = minimize(
        _bend_extend_2_tangent, # with tangent matching
        [0, 0, 0, 0, 0], 
        args=(
            curve_cps, 
            curve.length(),
            direction,
            list_to_c(target_tan0),  
            list_to_c(target_tan1), 
            70   # NOTE: Low values cause instable resutls
        ),
        method='L-BFGS-B',
    )
    if not out.success:
        print(f'Curve_match_tangents::WARNING::optimization not successfull')
        if flags.VERBOSE:
            print(out)

    shift = out.x

    fin_curve_cps = [
        curve_cps[0].tolist(),
        [curve_cps[1][0] + shift[0], curve_cps[1][1] + shift[1]], 
        [curve_cps[2][0] + shift[2], curve_cps[2][1] + shift[3]],
        (curve_cps[-1] + direction*shift[-1]).tolist(), 
    ]

    if return_as_edge:
        fin_inv_edge = CurveEdge(
            start=fin_curve_cps[0], 
            end=fin_curve_cps[-1], 
            control_points=fin_curve_cps[1:3],
            relative=False
        )
        return fin_inv_edge
    
    return fin_curve_cps


# ---- Utils ----

def _dist(v1, v2):
    return norm(v2-v1)



from copy import copy
from numpy.linalg import norm
import numpy as np

# Custom
from .edge_factory import EdgeSeqFactory
from .interface import Interface
from .generic_utils import close_enough
from . import flags

verbose=flags.VERBOSE

class StitchingRule():
    """High-level stitching instructions connecting two component interfaces
    """
    def __init__(self, int1:Interface, int2:Interface) -> None:
        """
        NOTE: When connecting interfaces with multiple edge count on both sides, 
            1) Note that the edge sequences may change their structure.
                Involved interfaces and corresponding patterns will be updated automatically
                Use of the same interfaces in other stitches (creating 3+way stitch edge) may fail. 
            2) The interfaces' edges are matched based on the provided order in the interface. 
            The order can be controlled at the moment of interface creation
        """
        self.int1 = int1
        self.int2 = int2

        if not self.isMatching():
            self.match_interfaces()

        if verbose and not close_enough(
                len1:=int1.projecting_edges().length(), 
                len2:=int2.projecting_edges().length(), 
                tol=0.3):   # NOTE = 3 mm
            print(
                f'{self.__class__.__name__}::Warning::Projected edges do not match in the stitch: \n'
                f'{len1}: {int1}\n{len2}: {int2}')
    

    def isMatching(self, tol=0.05):
        # if both the breakdown and relative partitioning is similar

        rev_frac1 = self.int1.edges.fractions()
        rev_frac1.reverse()

        return (len(self.int1) == len(self.int2) 
                and (np.allclose(self.int1.edges.fractions(), self.int2.edges.fractions(), atol=tol)
                    or np.allclose(rev_frac1, self.int2.edges.fractions(), atol=tol)
                )
        )


    def match_interfaces(self):
        """ Subdivide the interface edges on both sides s.t. they are matching 
            and can be safely connected
            (same number of edges on each side and same relative fractions)
        
            Serializable format does not natively support t-stitches, 
            so the longer edges needs to be broken down into matching segments
        """

        # Eval the fractions corresponding to every segments in the interfaces
        # Using projecting edges to match desired gather patterns
        frac1 = self.int1.projecting_edges(on_oriented=True).fractions()
        frac2 = self.int2.projecting_edges(on_oriented=True).fractions()
        min_frac = min(min(frac1), min(frac2))  # projection tolerance should not be larger than the smallest fraction

        self._match_to_fractions(self.int1, frac2, tol=min(1e-3, min_frac / 2))
        self._match_to_fractions(self.int2, frac1, tol=min(1e-3, min_frac / 2))


    def _match_to_fractions(self, inter:Interface, to_add, tol=1e-3):
        """Add the vertices at given location to the edge sequence in a given interface

        Parameters:
            * inter -- interface to modify
            * to_add -- the faractions of segements to be projected onto the edge sequence in the inter
            * tol -- the proximity of vertices when they can be regarded as the same vertex.  
                    NOTE: tol should be shorter then the smallest expected edge
        """

        # NOTE Edge sequences to subdivide might be disconnected 
        # (even belong to different panels), so we need to subdivide per edge

        # Go over the edges keeping track of their fractions
        add_id, in_id = 0, 0
        covered_init, covered_added = 0, 0
        total_len = inter.projecting_edges().length()

        while in_id < len(inter.edges) and add_id < len(to_add):
            # projected edges since they represent the stitch sizes
            next_init = covered_init + inter.projecting_edges()[in_id].length() / total_len
            next_added = covered_added + to_add[add_id]
            if close_enough(next_init, next_added, tol):
                # the vertex exists, skip
                in_id += 1
                add_id += 1
                covered_init, covered_added = next_init, next_added
            elif next_init < next_added:
                # add on the next step
                in_id += 1
                covered_init = next_init
            else:
                # add a vertex to the edge at the new location
                # Eval on projected edge
                in_frac = inter.projecting_edges()[in_id].length() / total_len
                new_v_loc = in_frac - (next_init - next_added)
                split_frac = new_v_loc / in_frac
                base_edge, base_panel = inter.edges[in_id], inter.panel[in_id]

                # Check edge orientation
                flip = inter.needsFlipping(in_id)
                if flip: 
                    split_frac = 1 - split_frac
                    if verbose:
                        print(f'{self.__class__.__name__}::INFO::{base_edge} from {base_panel.name} reoriented in interface')

                # Split the base edge accordingly
                subdiv = base_edge.subdivide_len([split_frac, 1 - split_frac])

                inter.panel[in_id].edges.substitute(base_edge, subdiv)  # Update the panel
                                                                        # Always follows the edge order in the panel
                # Swap subdiv order for interface to s.w. the interface sequence remains oriented
                if flip: 
                    subdiv.edges.reverse()
                    
                # Update interface accordingly
                inter.substitute(base_edge, subdiv, [inter.panel[in_id] for _ in range(len(subdiv))])

                # next step
                # By the size of new edge
                covered_init += inter.projecting_edges()[in_id].length() / total_len 
                covered_added = next_added
                in_id += 1
                add_id += 1

        if add_id != len(to_add):
            raise RuntimeError(f'{self.__class__.__name__}::Error::Projection failed')
                

    def assembly(self):
        """Produce a stitch that connects two interfaces
        """
        if verbose and not self.isMatching():
            print(f'{self.__class__.__name__}::WARNING::Stitch sides do not match on assembly!!')

        stitches = []

        for i, j in zip(range(len(self.int1.edges)), range(len(self.int2.edges))):
            stitches.append([
                {
                    'panel': self.int1.panel[i].name,  # corresponds to a name. 
                                            # Only one element of the first level is expected
                    'edge': self.int1.edges[i].geometric_id
                },
                {
                    'panel': self.int2.panel[j].name,
                    'edge': self.int2.edges[j].geometric_id
                }
            ])
        return stitches


class Stitches():
    """Describes a collection of StitchingRule objects
        Needed for more compact specification and evaluation of those rules
    """
    def __init__(self, *rules) -> None:
        """Rules -- any number of tuples of two interfaces (Interface, Interface) """

        self.rules = [StitchingRule(int1, int2) for int1, int2 in rules]

    def append(self, pair): 
        self.rules.append(StitchingRule(*pair))

    def __getitem__(self, id):
        return self.rules[id]

    def assembly(self):
        stitches = []
        for rule in self.rules:
            stitches += rule.assembly()
        return stitches


"""A Python library for building sewing patterns
"""

# Building blocks
from .component import Component
from .panel import Panel
from .edge import *
from .connector import Stitches
from .interface import Interface
from .edge_factory import EdgeSeqFactory as esf

# Operations
import pygarment.operators as ops
import pygarment.generic_utils as utils
import pygarment.flags as flags

# Parameter support
from .params import *




from copy import copy
from numpy.linalg import norm
import numpy as np

# Custom
from .edge import Edge, EdgeSequence
from .generic_utils import close_enough

class Interface():
    """Description of an interface of a panel or component
        that can be used in stitches as a single unit
    """
    def __init__(self, panel, edges, ruffle=1.):
        """
        Parameters:
            * panel - Panel object
            * edges - Edge or EdgeSequence -- edges in the panel that are allowed to connect to
            * ruffle - ruffle coefficient for a particular edge. Interface object will supply projecting_edges() shape
                s.t. the ruffles with the given rate are created. Default = 1. (no ruffles, smooth connection)
        """

        self.edges = edges if isinstance(edges, EdgeSequence) else EdgeSequence(edges)
        self.panel = [panel for _ in range(len(self.edges))]  # matches every edge 

        # Allow to enfoce change the direction of edge 
        # (used in many-to-many stitches correspondance determination)
        self.edges_flipping = [False for _ in range(len(self.edges))]

        # Ruffles are applied to sections
        # Since extending a chain of edges != extending each edge individually
        self.ruffle = [dict(coeff=ruffle, sec=[0, len(self.edges)])]

    def projecting_edges(self, on_oriented=False) -> EdgeSequence:
        """Return edges shape that should be used when projecting interface onto another panel
            NOTE: reflects current state of the edge object. Call this function again if egdes change (e.g. their direction)
        """
        # Per edge set ruffle application
        projected = self.edges.copy() if not on_oriented else self.oriented_edges()
        for r in self.ruffle:
            if not close_enough(r['coeff'], 1, 1e-3):
                projected[r['sec'][0]:r['sec'][1]].extend(1 / r['coeff'])
        
        return projected

    def needsFlipping(self, i):
        """ Check if particular edge should be re-oriented to follow the general direction of the interface
        """
        if self.edges_flipping[i]:
            return True
        
        # Otherwise, try to evaluate

        e = self.edges[i]
        panel = self.panel[i]
        s_3d, end_3d = panel.point_to_3D(e.start), panel.point_to_3D(e.end)

        # Corener cases
        if i == 0:
            next, next_panel = self.edges[(i+1 % len(self.edges))], self.panel[(i+1)  % len(self.edges)]
            next_3d = next_panel.point_to_3D(next.midpoint())

            # check by start vertex
            # NOTE this can misfire in particular 3D orentations
            return norm(s_3d - next_3d) < norm(end_3d - next_3d)
        if i == len(self.edges) - 1:
            prev, prev_panel = self.edges[i-1], self.panel[i-1]
            prev_3d = prev_panel.point_to_3D(prev.midpoint())

            # check by start vertex
            # NOTE this can misfire in particular 3D orentations
            return norm(s_3d - prev_3d) > norm(end_3d - prev_3d)

        # Mid case
        # Utilize distance from the end vertex to the next panel 
        # start -> prev + end -> next or other way around  
        prev, prev_panel = self.edges[i-1], self.panel[i-1]
        next, next_panel = self.edges[(i+1 % len(self.edges))], self.panel[(i+1)  % len(self.edges)]

        # Optimal order in 3D
        prev_3d = prev_panel.point_to_3D(prev.midpoint())
        next_3d = next_panel.point_to_3D(next.midpoint())

        forward_order_dist = norm(s_3d - prev_3d) + norm(end_3d - next_3d)
        flipped_order_dist = norm(s_3d - next_3d) + norm(end_3d - prev_3d)

        return flipped_order_dist < forward_order_dist

    # ANCHOR --- Info ----
    def oriented_edges(self):
        """ Orient the edges withing the interface sequence along the general direction of the interface

            Creates a copy of the interface s.t. not to disturb the original edge objects
        """
        # NOTE we cannot we do the same for the edge sub-sequences:
        #  - midpoint of a sequence is less representative
        #  - more likely to have weird relative 3D orientations
        # => heuristic won't work as well

        oriented = self.edges.copy()

        for i in range(len(self.edges)):
            if self.needsFlipping(i):
                oriented[i].reverse()
                oriented[i].flipped = True
            else:
                oriented[i].flipped = False
        return oriented

    def verts_3d(self):
        """Return 3D locations of all vertices that participate in the interface"""

        verts_2d = []
        matching_panels = []
        for e, panel in zip(self.edges, self.panel):
            if all(e.start is not v for v in verts_2d):  # Ensuring uniqueness
                verts_2d.append(e.start)
                matching_panels.append(panel)
            
            if all(e.end is not v for v in verts_2d):  # Ensuring uniqueness
                verts_2d.append(e.end)
                matching_panels.append(panel)

        # To 3D
        verts_3d = []
        for v, panel in zip(verts_2d, matching_panels):
            verts_3d.append(panel.point_to_3D(v))

        return np.asarray(verts_3d)

    def bbox_3d(self):
        """Return Interface bounding box"""
        verts = self.verts_3d()

        return [np.min(verts, axis=0), np.max(verts, axis=0)]

    def __len__(self):
        return len(self.edges)
    
    def __str__(self) -> str:
        return f'Interface: {[p.name for p in self.panel]}: {str(self.oriented_edges())}'
    
    def __repr__(self) -> str:
        return self.__str__()

    # ANCHOR --- Interface Updates -----

    def reverse(self, with_edge_dir_reverse=False):
        """Reverse the order of edges in the interface
            (without updating the edge objects)

            Reversal is useful for reordering interface edges for correct matching in the multi-stitches
        """
        self.edges.edges.reverse()  
        self.panel.reverse()
        self.edges_flipping.reverse()
        if with_edge_dir_reverse:
            self.edges_flipping = [not e for e in self.edges_flipping]

        enum = len(self.edges)
        for r in self.ruffle:
            # Update ids
            r['sec'][0] = enum - r['sec'][0]
            r['sec'][1] = enum - r['sec'][1]
            # Swap
            r['sec'][0], r['sec'][1] = r['sec'][1], r['sec'][0]
        
        return self

    def reorder(self, curr_edge_ids, projected_edge_ids):
        """Change the order of edges from curr_edge_ids to projected_edge_ids in the interface

            Note that the input should prescrive new ordering for all affected edges
            e.g. if moving 0 -> 1, specify the new location for 1 as well
        """
        
        for i, j in zip(curr_edge_ids, projected_edge_ids):
            for r in self.ruffle:
                if (i >= r['sec'][0] and i < r['sec'][1] 
                        and (j < r['sec'][0] or j >= r['sec'][1])):
                    raise NotImplementedError(
                        f'{self.__class__.__name__}::Error::reordering between panel-related sub-segments is not supported')
        
        new_edges = EdgeSequence()
        new_panel_list = []
        new_flipping_info = []
        for i in range(len(self.panel)):
            id = i if i not in curr_edge_ids else projected_edge_ids[curr_edge_ids.index(i)]
            # edges
            new_edges.append(self.edges[id])
            new_flipping_info.append(self.edges_flipping[id])
            # panels
            new_panel_list.append(self.panel[id])
            
        self.edges = new_edges
        self.panel = new_panel_list
        self.edges_flipping = new_flipping_info

    def substitute(self, orig, new_edges, new_panels):
        """Update the interface edges with correct correction of panels
            * orig -- could be an edge object or the id of edges that need substitution
            * new_edges -- new edges to insert in place of orig
            * new_panels -- per-edge panel objects indicating where each of new_edges belong to
        
        NOTE: the ruffle indicator for the new_edges is expected to be the same as for orig edge
        Specifying new indicators is not yet supported

        """
        if isinstance(orig, Edge):
            orig = self.edges.index(orig)
        if orig < 0: 
            orig = len(self.edges) + orig 
        self.edges.substitute(orig, new_edges)

        # Update panels & flip info
        self.panel.pop(orig)
        self.edges_flipping.pop(orig)
        if isinstance(new_panels, list) or isinstance(new_panels, tuple):
            for j in range(len(new_panels)):
                self.panel.insert(orig + j, new_panels[j])
                self.edges_flipping.insert(orig + j, False)
        else: 
            self.panel.insert(orig, new_panels)
            self.edges_flipping.insert(orig, False)

        # Propagate ruffle indicators
        ins_len = 1 if isinstance(new_edges, Edge) else len(new_edges)
        if ins_len > 1:
            for it in self.ruffle:  # UPD ruffle indicators
                if it['sec'][0] > orig:
                    it['sec'][0] += ins_len - 1
                if it['sec'][1] > orig:
                    it['sec'][1] += ins_len - 1

        return self

    # ANCHOR ----- Statics ----
    @staticmethod
    def from_multiple(*ints):
        """Create interface from other interfaces: 
            * Allows to use different panels in one interface
            * different ruffle values in one interface
            
            # NOTE the relative order of edges is preserved from the
            original interfaces and the incoming interface sequence
            This order will then be used in the SrtitchingRule when
            determing connectivity between interfaces
        """
        new_int = copy(ints[0])  # shallow copy -- don't create unnecessary objects
        new_int.edges = EdgeSequence()
        new_int.edges_flipping = []
        new_int.panel = []
        new_int.ruffle = []
        
        for elem in ints:
            shift = len(new_int.edges)
            new_int.ruffle += [copy(r) for r in elem.ruffle]
            for r in new_int.ruffle[-len(elem.ruffle):]:
                r.update(sec=[r['sec'][0] + shift, r['sec'][1] + shift])

            new_int.edges.append(elem.edges)
            new_int.panel += elem.panel 
            new_int.edges_flipping += elem.edges_flipping 
            
        return new_int 

    @staticmethod
    def _is_order_matching(panel_s, vert_s, panel_1, vert1, panel_2, vert2) -> bool:
        """Check which of the two vertices vert1 (panel_1) or vert2 (panel_2) is closer to the vert_s 
            from panel_s in 3D"""
        s_3d = panel_s.point_to_3D(vert_s)
        v1_3d = panel_1.point_to_3D(vert1)
        v2_3d = panel_2.point_to_3D(vert2)

        return norm(v1_3d - s_3d) < norm(v2_3d - s_3d)




import numpy as np
from copy import copy
from argparse import Namespace
from scipy.spatial.transform import Rotation as R

# Custom
from pattern.core import BasicPattern
from pattern.wrappers import VisPattern
from .base import BaseComponent
from .edge import Edge, EdgeSequence
from .connector import Stitches
from .generic_utils import close_enough, vector_align_3D

class Panel(BaseComponent):
    """ A Base class for defining a Garment component corresponding to a single flat fiece of fabric
    
    Defined as a collection of edges on a 2D grid with specified 3D placement (world coordinates)
    
    NOTE: All operations methods return 'self' object to allow sequential applications

    """
    def __init__(self, name) -> None:
        super().__init__(name)

        self.translation = np.zeros(3)
        self.rotation = R.from_euler('XYZ', [0, 0, 0])  # zero rotation
        # NOTE: initiating with empty sequence allows .append() to it safely
        self.edges =  EdgeSequence() 

    # Info
    def pivot_3D(self):
        """Pivot point of a panel in 3D"""
        return self.point_to_3D([0, 0])

    # ANCHOR - Operations -- update object in-place 
    def set_pivot(self, point_2d, replicate_placement=False):
        """Specify 2D point w.r.t. panel local space
            to be used as pivot for translation and rotation

        Parameters:
            * point_2d -- desired point 2D point w.r.t current pivot (origin) of panel local space
            * replicate_placement -- will replicate the location of the panel as it was before pivot change
                default - False (no adjustment, the panel may "jump" in 3D)
        """
        point_2d = copy(point_2d)  # Remove unwanted object reference 
                                   # In case an actual vertex was used as a target point

        if replicate_placement:
            self.translation = self.point_to_3D(point_2d)

        # UPD vertex locations relative to new pivot
        for v in self.edges.verts():
            v[0] -= int(point_2d[0])
            v[1] -= int(point_2d[1])

    def top_center_pivot(self):
        """One of the most useful pivots 
            is the one in the middle of the top edge of the panel 
        """
        vertices = np.asarray(self.edges.verts())

        # out of 2D bounding box sides' midpoints choose the one that is highest in 3D
        top_right = vertices.max(axis=0)
        low_left = vertices.min(axis=0)
        mid_x = (top_right[0] + low_left[0]) / 2
        mid_y = (top_right[1] + low_left[1]) / 2
        mid_points_2D = [
            [mid_x, top_right[1]], 
            [mid_x, low_left[1]],
            [top_right[0], mid_y],
            [low_left[0], mid_y]
        ]
        mid_points_3D = np.vstack(tuple(
            [self.point_to_3D(coords) for coords in mid_points_2D]
        ))
        top_mid_point = mid_points_3D[:, 1].argmax()

        self.set_pivot(mid_points_2D[top_mid_point])

        return self

    def translate_by(self, delta_vector):
        """Translate panel by a vector"""
        self.translation = self.translation + np.array(delta_vector)
        self.autonorm()

        return self
    
    def translate_to(self, new_translation):
        """Set panel translation to be exactly that vector"""
        self.translation = np.asarray(new_translation)
        self.autonorm()

        return self
    
    def rotate_by(self, delta_rotation: R):
        """Rotate panel by a given rotation
            * delta_rotation: scipy rotation object
        """
        self.rotation = delta_rotation * self.rotation
        self.autonorm()

        return self

    def rotate_to(self, new_rot: R):
        """Set panel rotation to be exactly the given rotation
            * new_rot: scipy rotation object
        """   
        if not isinstance(new_rot, R):
            raise ValueError(f'{self.__class__.__name__}::Error::Only accepting rotations in scipy format')
        self.rotation = new_rot
        self.autonorm()

        return self

    def rotate_align(self, vector):
        """Set panel rotation s.t. it's norm is aligned with a given 3D vector"""

        vector = np.asarray(vector)
        vector = vector / np.linalg.norm(vector)
        n = self.norm()
        self.rotate_by(vector_align_3D(n, vector))

        return self

    def center_x(self):
        """Adjust translation over x s.t. the center of the panel is aligned with the Y axis (center of the body)"""

        center_3d = self.point_to_3D(self._center_2D())
        self.translation[0] += -center_3d[0]

        return self

    def autonorm(self):
        """Update right/wrong side orientation, s.t. the normal of the surface looks outside of the world origin, 
            taking into account the shape and the global position.
        
            This should provide correct panel orientation in most cases.

            NOTE: for best results, call autonorm after translation specification
        """
        norm_dr = self.norm()
        
        # NOTE: Nothing happens if self.translation is zero
        if np.dot(norm_dr, self.translation) < 0: 
            # Swap if wrong  
            self.edges.reverse()
        
        return self

    def mirror(self, axis=[0, 1]):
        """Swap this panel with it's mirror image
        
            Axis specifies 2D axis to swap around: Y axis by default
        """

        # Case Around Y
        if close_enough(axis[0], tol=1e-4):  # reflection around Y

            # Vertices
            self.edges.reflect([0, 0], [0, 1])
            
            # Position
            self.translation[0] *= -1

            # Rotations
            curr_euler = self.rotation.as_euler('XYZ')
            curr_euler[1] *= -1  
            curr_euler[2] *= -1  
            self.rotate_to(R.from_euler('XYZ', curr_euler))  

            # Fix right/wrong side
            self.autonorm()
        else:
            raise NotImplementedError(f'{self.name}::Error::Mirrowing over arbitrary axis is not implemented')

        return self
        
    # ANCHOR - Build the panel -- get serializable representation
    def assembly(self):
        """Convert panel into serialazable representation
        
         # SIM Note that Qualoth simulator does not support internal loops in panels,
            hence panel EdgeSequence is assumed to be a single loop of edges
        """
        # always start from zero for consistency between panels
        self.set_pivot(self.edges[0].start, replicate_placement=True)

        # Basics
        panel = Namespace(
            translation=self.translation.tolist(),
            rotation=self.rotation.as_euler('XYZ', degrees=True).tolist(), 
            vertices=[self.edges[0].start], 
            edges=[])

        for i in range(len(self.edges)):
            vertices, edge = self.edges[i].assembly()

            # add new vertices
            if panel.vertices[-1] == vertices[0]:   # We care if both point to the same vertex location, not necessarily the same vertex object
                vert_shift = len(panel.vertices) - 1  # first edge vertex = last vertex already in the loop
                panel.vertices += vertices[1:] 
            else: 
                vert_shift = len(panel.vertices)
                panel.vertices += vertices

            # upd vertex references in edges according to new vertex ids in 
            # the panel vertex loop
            edge['endpoints'] = [id + vert_shift for id in edge['endpoints']]
                
            edge_shift = len(panel.edges)  # before adding new ones
            self.edges[i].geometric_id = edge_shift   # remember the mapping of logical edge to geometric id in panel loop
            panel.edges.append(edge)

        # Check closing of the loop and upd vertex reference for the last edge
        if panel.vertices[-1] == panel.vertices[0]:
            panel.vertices.pop()
            panel.edges[-1]['endpoints'][-1] = 0

        spattern = BasicPattern()
        spattern.name = self.name
        spattern.pattern['panels'] = {self.name: vars(panel)}

        # Assembly stitching info (panel might have inner stitches)
        spattern.pattern['stitches'] = self.stitching_rules.assembly()
            
        return spattern

    # ANCHOR utils
    def _center_2D(self):
        """Approximate Location of the panel center. 
            
            NOTE: uses crude linear approximation for curved edges
        """
        # NOTE: assuming that edges are organized in a loop and share vertices
        lin_edges = EdgeSequence([e.linearize() for e in self.edges])
        verts = lin_edges.verts()

        return np.mean(verts, axis=0)

    def point_to_3D(self, point_2d):
        """Calculate 3D location of a point given in the local 2D plane """
        point_2d = np.asarray(point_2d)
        if len(point_2d) == 2:
            point_2d = np.append(point_2d, 0)

        point_3d = self.rotation.apply(point_2d)
        point_3d += self.translation

        return point_3d

    def norm(self):
        """Normal direction for the current panel"""

        # Take linear version of the edges
        # To correctly process edges with extreme curvatures
        lin_edges = EdgeSequence([e.linearize() for e in self.edges])

        # center of mass
        verts = lin_edges.verts()
        center = np.mean(verts, axis=0)
        center_3d = self.point_to_3D(center)

        # To make norm evaluation work for non-convex panels
        # Evalute norm candidates for all edges and then weight them. 
        # The dominant norm direction should be the correct one 
        norms = []
        for e in lin_edges:
            vert_0 = self.point_to_3D(e.start)
            vert_1 = self.point_to_3D(e.end)

            # Pylance + NP error for unreachanble code -- see https://github.com/numpy/numpy/issues/22146
            # Works ok for numpy 1.23.4+
            norm = np.cross(vert_1 - vert_0, center_3d - vert_0)
            norm /= np.linalg.norm(norm) 
            norms.append(norm)

        # Current norm direction
        avg_norm = sum(norms) / len(norms)

        if close_enough(np.linalg.norm(avg_norm), 0):
            # Indecisive averaging, so using just one of the norms
            # NOTE: sometimes happens on thin arcs
            avg_norm = norms[0]   
            print(f'{self.__class__.__name__}::{self.name}::WARNING::Norm evaluation failed, assigning norm based on the first edge')
        return avg_norm / np.linalg.norm(avg_norm)

    def bbox3D(self):
        """Evaluate 3D bounding box of the current panel"""

        # Using curve linearization for more accurate approximation of bbox
        lin_edges = EdgeSequence([e.linearize() for e in self.edges])
        verts_2d = lin_edges.verts()
        verts_3d = np.asarray([self.point_to_3D(v) for v in verts_2d])

        return verts_3d.min(axis=0), verts_3d.max(axis=0)




import numpy as np
from numpy.linalg import norm
import svgpathtools as svgpath
from scipy.optimize import minimize

# Custom
from .edge import EdgeSequence, Edge, CurveEdge
from .generic_utils import close_enough, c_to_list, list_to_c
from . import flags

class EdgeSeqFactory:
    """Create EdgeSequence objects for some common edge seqeunce patterns
    """

    @staticmethod
    def from_verts(*verts, loop=False):
        """Generate edge sequence from given vertices. If loop==True, the method also closes the edge sequence as a loop
        """

        seq = EdgeSequence(Edge(verts[0], verts[1]))
        for i in range(2, len(verts)):
            seq.append(Edge(seq[-1].end, verts[i]))

        if loop:
            seq.append(Edge(seq[-1].end, seq[0].start))
        
        seq.isChained()
        return seq

    @staticmethod
    def from_fractions(start, end, frac=[1]):
        """A sequence of edges between start and end wich lengths are distributed
            as specified in frac list 
        Parameters:
            * frac -- list of legth fractions. Every entry is in (0, 1], 
                all entries sums up to 1
        """
        frac = [abs(f) for f in frac]
        if not close_enough(fsum:=sum(frac), 1, 1e-4):
            raise RuntimeError(f'EdgeSequence::Error::fraction is incorrect. The sum {fsum} is not 1')

        vec = np.asarray(end) - np.asarray(start)
        verts = [start]
        for i in range(len(frac) - 1):
            verts.append(
                [verts[-1][0] + frac[i]*vec[0],
                verts[-1][1] + frac[i]*vec[1]]
            )
        verts.append(end)
        
        return EdgeSeqFactory.from_verts(*verts)

    @staticmethod
    def side_with_cut(start=(0,0), end=(1,0), start_cut=0, end_cut=0):
        """ Edge with internal vertices that allows to stitch only part of the border represented
            by the long side edge

            start_cut and end_cut specify the fraction of the edge to to add extra vertices at
        """

        nstart, nend = np.array(start), np.array(end)
        verts = [start]

        if start_cut > 0:
            verts.append((start + start_cut * (nend-nstart)).tolist())
        if end_cut > 0:
            verts.append((end - end_cut * (nend-nstart)).tolist())
        verts.append(end)

        edges = EdgeSeqFactory.from_verts(*verts)

        return edges

    @staticmethod
    def dart_shape(width, side_len=None, depth=None):
        """Shape of simple triangular dart: 
            specified by desired width and either the dart side length or depth
        """

        if side_len is None and depth is None:
            raise ValueError(
                'EdgeFactory::Error::dart shape is not fully specified.'
                ' Add dart side length or dart perpendicular'
            )

        if depth is None:
            if width / 2 > side_len: 
                raise ValueError(
                    f'EdgeFactory::Error::Requested dart shape (w={width}, side={side_len}) '
                    'does not form a valid triangle')
            depth = np.sqrt((side_len**2 - (width / 2)**2))

        return EdgeSeqFactory.from_verts([0, 0], [width / 2, -depth], [width, 0])

    @staticmethod
    def curve_from_extreme(start, end, target):
        """Create (Quadratic) curve edge that 
            has an extreme point as close as possible to target_extreme
            with extreme point aligned with it
        """
        rel_target = _abs_to_rel_2d(start, end, target)

        out = minimize(
            _fit_y_extremum, 
            rel_target[1],    
            args=(rel_target)
        )

        if not out.success:
            print('Curve From Extreme::WARNING::Optimization not successful')
            if flags.VERBOSE:
                print(out)

        cp = [rel_target[0], out.x.item()]

        return CurveEdge(start, end, control_points=[cp], relative=True)
    
    @staticmethod
    def curve_3_points(start, end, target):
        """Create (Quadratic) curve edge between start and end that
            passes through the target point 
        """
        rel_target = _abs_to_rel_2d(start, end, target)

        if rel_target[0] > 1 or rel_target[0] < 0:
            raise NotImplementedError(
                f"EdgeFactory::Curve_by_3_points::ERROR::requested target point's projection "
                "is outside of the base edge, which is not yet supported"
            )

        # Initialization with a target point as control point
        # Ensures very smooth, minimal solution
        out = minimize(
            _fit_pass_point, 
            rel_target,    
            args=(rel_target)
        )

        if not out.success:
            print('Curve From Extreme::WARNING::Optimization not successful')
            if flags.VERBOSE:
                print(out)

        cp = out.x.tolist()

        return CurveEdge(start, end, control_points=[cp], relative=True)

    @staticmethod
    def curve_from_len_tangents(start, end, cp, target_len, target_tan0, target_tan1):
        """Find a curve with given relative curvature parameters"""

# Utils
def _abs_to_rel_2d(start, end, point):
    """Convert control points coordinates from absolute to relative"""
    start, end = np.asarray(start), np.asarray(end)
    edge = end - start
    edge_len = norm(edge)

    point_vec = np.asarray(point) - start

    converted = [None, None]
    # X
    # project control_vec on edge by dot product properties
    projected_len = edge.dot(point_vec) / edge_len 
    converted[0] = projected_len / edge_len
    # Y
    control_projected = edge * converted[0]
    vert_comp = point_vec - control_projected  
    converted[1] = norm(vert_comp) / edge_len

    # Distinguish left&right curvature
    converted[1] *= -np.sign(np.cross(point_vec, edge)) 
    
    return np.asarray(converted)

def _extreme_points(curve, on_x=False, on_y=True):
    """Return extreme points of the input curve
        NOTE: this does NOT include the border vertices of an edge
    """
    # Variation of https://github.com/mathandy/svgpathtools/blob/5c73056420386753890712170da602493aad1860/svgpathtools/bezier.py#L197
    poly = svgpath.bezier2polynomial(curve, return_poly1d=True)
    
    x_extremizers, y_extremizers = [], []
    if on_y:
        y = svgpath.imag(poly)
        dy = y.deriv()
        
        y_extremizers = svgpath.polyroots(dy, realroots=True,
                                            condition=lambda r: 0 < r < 1)
    if on_x:
        x = svgpath.real(poly)
        dx = x.deriv()
        x_extremizers = svgpath.polyroots(dx, realroots=True,
                                    condition=lambda r: 0 < r < 1)
    all_extremizers = x_extremizers + y_extremizers

    extreme_points = np.array([c_to_list(curve.point(t)) for t in all_extremizers])

    return extreme_points

def _fit_y_extremum(cp_y, target_location):
    """ Fit the control point of basic [[0, 0] -> [1, 0]] Quadratic Bezier s.t. 
        it's expremum is close to target location.

        * cp_y - initial guess for Quadratic Bezier control point y coordinate
            (relative to the edge)
        * target_location -- target to fit extremum to -- 
            expressed in RELATIVE coordinates to your desired edge
    """

    control_bezier = np.array([
        [0, 0], 
        [target_location[0], cp_y[0]], 
        [1, 0]
    ])
    params = list_to_c(control_bezier)
    curve = svgpath.QuadraticBezier(*params)

    extremum = _extreme_points(curve)

    if not len(extremum):
        raise RuntimeError('No extreme points!!')

    diff = np.linalg.norm(extremum - target_location)

    return diff**2 

def _fit_pass_point(cp, target_location):
    """ Fit the control point of basic [[0, 0] -> [1, 0]] Quadratic Bezier s.t. 
        it's expremum is close to target location.

        * cp_y - initial guess for Quadratic Bezier control point y coordinate
            (relative to the edge)
        * target_location -- target to fit extremum to -- 
            expressed in RELATIVE coordinates to your desired edge
    """
    control_bezier = np.array([
        [0, 0], 
        cp, 
        [1, 0]
    ])
    params = list_to_c(control_bezier)
    curve = svgpath.QuadraticBezier(*params)

    inter_segment = svgpath.Line(
            target_location[0] + 1j * target_location[1] * 2,
            target_location[0] + 1j * (- target_location[1] * 2)
        )

    intersect_t = curve.intersect(inter_segment)
    point = curve.point(intersect_t[0][0])

    diff = abs(point - list_to_c(target_location))

    return diff**2 




from typing import Any
from .connector import Stitches

class BaseComponent():
    """Basic interface for garment-related components
    
        NOTE: modifier methods return self object to allow chaining of the operations
    """

    def __init__(self, name) -> None:
        self.name = name

        # List or dictionary of the interfaces of this components
        # available for connectivity with other components
        self.interfaces = {}

        # Rules for connecting subcomponents
        self.stitching_rules = Stitches()

    # Info
    def pivot_3D(self):
        """Pivot location of a component in 3D"""
        return [0, 0, 0]
    
    def bbox(seld):
        """Bounding box"""
        return 0, 0, 0, 0, 0, 0

    # Operations
    def translate_by(self, delta_translation):
        return self

    def translate_to(self, new_translation):
        """Set panel translation to be exactly that vector"""
        return self

    def rotate_by(self, delta_rotation):
        return self
    
    def rotate_to(self, new_rot):
        return self

    def assembly(self, *args,**kwds):
        return {}

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return self.assembly(*args,**kwds)


"""Parameter class wrappers around parameter files allowing definition of computed parameters
"""

import yaml
from pathlib import Path


class BodyParametrizationBase():
    """Base class for body parametrization wrappers that allows definition of 
        dependent parameters
    """

    def __init__(self, param_file='') -> None:
        
        self.params = {}
        self.load(param_file)

    def __getitem__(self, key):
        return self.params[key]
    
    def __iter__(self):
        """
        Return an iterator of dict keys
        """
        return iter(self.params)
    
    # Updates
    def __setitem__(self, key, value):
        self.params[key] = value
        self.eval_dependencies()

    def load(self, param_file):
        """Load new values from file"""
        with open(param_file, 'r') as f:
            dict = yaml.safe_load(f)['body']
        self.params.update(dict)
        self.eval_dependencies()  # Parameters have been updated

    # Processing
    def eval_dependencies(self):
        """Evaluate dependent attributes, e.g. after a new value has been set
        
            Define your dependent parameters in the overload of this function
        """
        pass
        
    # Save
    def save(self, path, name='body_measurements'):
        with open(Path(path) / f'{name}.yaml', 'w') as f:
            yaml.dump(
                {'body': self.params}, 
                f,
                default_flow_style=False
            )


"""
    Simple Rotation Conversion routines (Maya-Python2.7-Compatible!!)
    TODO: Can be substituted with scipy rotation transformation routines for Maya2022+
"""
import numpy as np
import math as m
import sys


# Thanks to https://www.meccanismocomplesso.org/en/3d-rotations-and-euler-angles-in-python/ for the code
def _Rx(theta):
    return np.matrix([
        [1, 0           , 0           ],
        [0, m.cos(theta), -m.sin(theta)],
        [0, m.sin(theta), m.cos(theta)]])


def _Ry(theta):
    return np.matrix([
        [m.cos(theta), 0, m.sin(theta)],
        [0           , 1, 0           ],
        [-m.sin(theta), 0, m.cos(theta)]])


def _Rz(theta):
    return np.matrix([
        [m.cos(theta), -m.sin(theta), 0],
        [m.sin(theta), m.cos(theta) , 0],
        [0           , 0            , 1]])


def euler_xyz_to_R(euler):
    """Convert to Rotation matrix.
        Expects input in degrees.
        Only support Maya convension of intrinsic xyz Euler Angles
    """
    return _Rz(np.deg2rad(euler[2])) * _Ry(np.deg2rad(euler[1])) * _Rx(np.deg2rad(euler[0]))


def R_to_euler(R):
    """
        Convert Rotation matrix to Euler-angles in degrees (in Maya convension of intrinsic xyz Euler Angles)
        NOTE: 
            Routine produces one of the possible Euler angles, corresponding to input rotations (the Euler angles are not uniquely defined)
    """
    tol = sys.float_info.epsilon * 10
  
    if abs(R[0, 0]) < tol and abs(R[1, 0]) < tol:
        eul1 = 0
        eul2 = m.atan2(-R[2, 0], R[0, 0])
        eul3 = m.atan2(-R[1, 2], R[1, 1])
    else:   
        eul1 = m.atan2(R[1, 0], R[0, 0])
        sp = m.sin(eul1)
        cp = m.cos(eul1)
        eul2 = m.atan2(-R[2, 0], cp * R[0, 0] + sp * R[1, 0])
        eul3 = m.atan2(sp * R[0, 2] - cp * R[1, 2], cp * R[1, 1] - sp * R[0, 1])
    
    return [np.rad2deg(eul3), np.rad2deg(eul2), np.rad2deg(eul1)]




from copy import deepcopy, copy
import numpy as np
from numpy.linalg import norm
import svgpathtools as svgpath  # https://github.com/mathandy/svgpathtools

# Custom
from .generic_utils import vector_angle, R2D, close_enough, c_to_list, list_to_c
from .flags import VERBOSE

class Edge():
    """Edge -- an individual segement of a panel border connecting two panel vertices, 
     the basic building block of panels

    Edges are defined on 2D coordinate system with Start vertex as an origin and (End-Start) as Ox axis
    """

    def __init__(self, start=[0,0], end=[0,0]) -> None:
        """ Simple edge inititalization.
        Parameters: 
            * start, end: from/to vertcies that the edge connectes, describing the _interface_ of an edge
        """

        assert not all(close_enough(s, e) for s, e in zip(start, end)), 'Start and end of an edge should differ'

        self.start = start  # NOTE: careful with references to vertex objects
        self.end = end

        # ID w.r.t. other edges in a super-panel
        # Filled out at the panel assembly time
        self.geometric_id = 0

    # Info
    def length(self):
        """Return current length of an edge.
            Since vertices may change their locations externally, the length is dynamically evaluated
        """
        return self._straight_len()

    def _straight_len(self):
        """Length of the edge ignoring the curvature"""
        return norm(np.asarray(self.end) - np.asarray(self.start))

    def __eq__(self, __o: object, tol=1e-2) -> bool:
        """Special implementation of comparison: same edges == edges can be connected by flat stitch
            Edges are the same if their length is the same (if their flattened representation is the same)
                => vertices do not have to be on the same locations

            NOTE: The edges may not have the same curvature and still be considered equal ("connectible")
        """

        if not isinstance(__o, Edge):
            return False

        # Base length is the same
        if close_enough(self.length(), __o.length(), tol=tol):
            return False

        return True

    def __str__(self) -> str:
        return f'Straight:[{self.start[0]:.2f}, {self.start[1]:.2f}]->[{self.end[0]:.2f}, {self.end[1]:.2f}]'

    def __repr__(self) -> str:
        """ 'Official string representation' -- for nice printing of lists of edges
        
        https://stackoverflow.com/questions/3558474/how-to-apply-str-function-when-printing-a-list-of-objects-in-python
        """
        return self.__str__()

    def midpoint(self):
        """Center of the edge"""
        return (np.array(self.start) + np.array(self.end)) / 2

    def shortcut(self):
        """Return straight shortcut for an edge, 
            as np.array
        
            For straight edges it's the same as the edge itself
        """

        return np.array([self.start, self.end])

    # Representation
    def as_curve(self):
        """As svgpath curve object"""
        # Get the nodes correcly
        nodes = np.vstack((self.start, self.end))

        params = nodes[:, 0] + 1j*nodes[:, 1]

        return svgpath.Line(*params)

    def linearize(self):
        """Return a linear approximation of an edge using the same vertex objects
        
            # NOTE: for the linear edge it is an egde
        """

        return self

    # Support for relative coordinate system
    def _rel_to_abs_2d(self, point):
        """Convert coordinates expressed relative to an edge into absolute """
        start, end = np.array(self.start), np.array(self.end)
        edge = end - start
        edge_perp = np.array([-edge[1], edge[0]])

        conv_start = self.start + point[0] * edge
        conv_point = conv_start + point[1] * edge_perp
        
        return conv_point

    def _abs_to_rel_2d(self, point):
        """Convert control points coordinates from absolute to relative"""
        start, end = np.array(self.start), np.array(self.end)
        edge = end - start
        edge_len = norm(edge)

        point_vec = np.asarray(point) - start

        converted = [None, None]
        # X
        # project control_vec on edge by dot product properties
        projected_len = edge.dot(point_vec) / edge_len 
        converted[0] = projected_len / edge_len
        # Y
        control_projected = edge * converted[0]
        vert_comp = point_vec - control_projected  
        converted[1] = norm(vert_comp) / edge_len

        # Distinguish left&right curvature
        converted[1] *= -np.sign(np.cross(point_vec, edge)) 
        
        return np.asarray(converted)

    # Actions
    def reverse(self):
        """Flip the direction of the edge"""
        self.start, self.end = self.end, self.start

        return self
    
    def reflect_features(self):
        """Reflect edge fetures from one side of the edge to the other"""
        # Nothing to do for straight edge
        return self

    def snap_to(self, new_start=[0, 0]):
        """Translate the edge vertices s.t. the start is at new_start
        """
        self.end[0] = self.end[0] - self.start[0] + new_start[0]
        self.end[1] = self.end[1] - self.start[1] + new_start[1]

        self.start[:] = new_start
        return self

    def rotate(self, angle):
        """Rotate edge by angle in place, using first point as a reference

        Parameters: 
            angle -- desired rotation angle in radians (!)
        """
        curr_start = copy(self.start)
        
        # set the start point to zero
        self.snap_to([0, 0])
        self.end[:] = np.matmul(R2D(angle), self.end)
        
        # recover the original location
        self.snap_to(curr_start)

        return self
        
    def subdivide_len(self, fractions: list, connect_internal_verts=True):
        """Add intermediate vertices to an edge, 
            splitting it's length according to fractions
            while preserving the overall shape

            * merge_internal -- if False, the newly inserted vertices would be defined
                as independent objects for each edge. If True, vertex objects will be shared
        """
        # Parametrized by length
        new_edges = self._subdivide(fractions, by_length=True)

        if connect_internal_verts:
            self._merge_subdiv_vertices(new_edges)
        
        return new_edges
    
    def subdivide_param(self, fractions: list, connect_internal_verts=True):
        """Add intermediate vertices to an edge, 
            splitting its curve parametrization according to fractions
            while preserving the overall shape

            NOTE: for line, it's the same as subdivision by length
        """
        
        new_edges = self._subdivide(fractions, by_length=False)

        if connect_internal_verts:
            self._merge_subdiv_vertices(new_edges)
        
        return new_edges


    def _subdivide(self, fractions: list, by_length=True):
        """Subdivide edge by length or curve parametrization

            NOTE: equivalent for straight lines
        """

        frac = [abs(f) for f in fractions]
        if not close_enough(fsum:=sum(frac), 1, 1e-4):
            raise RuntimeError(f'Edge Subdivision::Error::fraction is incorrect. The sum {fsum} is not 1')

        vec = np.asarray(self.end) - np.asarray(self.start)
        verts = [self.start]
        seq = EdgeSequence()
        for i in range(len(frac) - 1):
            verts.append(
                [verts[-1][0] + frac[i]*vec[0],
                verts[-1][1] + frac[i]*vec[1]]
            )
            seq.append(Edge(verts[-2], verts[-1]))
        verts.append(self.end)
        seq.append(Edge(verts[-2], verts[-1]))
        
        return seq

    def _merge_subdiv_vertices(self, subdivision):
        """Merge the vertices from cosecutive edges in the given edge subdivision"""

        for i in range(1, len(subdivision)):
            subdivision[i].start = subdivision[i-1].end
        return subdivision

    # Assembly into serializable object
    def assembly(self):
        """Returns the dict-based representation of edges, 
            compatible with core -> BasePattern JSON (dict) 
        """

        return [self.start, self.end], {"endpoints": [0, 1]}


class CircleEdge(Edge):
    """Curvy edge as circular arc"""

    def __init__(self, start=[0, 0], end=[0, 0], cy=None) -> None:
        """
            Define a circular arc edge
            * start, end: from/to vertcies that the edge connectes
            * cy: third point on a circle arc (= control point). 
                Expressed relatively w.r.t. distance between start and end. 
                X value for control point is fixed at x=0.5 (edge center) to avoid ambiguity
            
            NOTE: represening control point in relative coordinates
            allows preservation of curvature (arc angle, relative raidus w.r.t. straight edge length)
            When distance between vertices shrinks / extends

            NOTE: full circle not supported: start & end should differ
        """
        super().__init__(start, end)
        
        self.control_y = cy

    def length(self):
        """Return current length of an edge.
            Since vertices may change their locations externally, the length is dynamically evaluated
        """
        return self._rel_radius() * self._straight_len() * self._arc_angle()

    def __str__(self) -> str:

        points = [self.start, [0.5, self.control_y]]

        str = [f'[{p[0]:.2f}, {p[1]:.2f}]->' for p in points]
        str += [f'[{self.end[0]:.2f}, {self.end[1]:.2f}]']

        return 'Arc:' + ''.join(str)
    
    def midpoint(self):
        """Center of the edge"""
        return self._rel_to_abs_2d([0.5, self.control_y])

    # Actions
    def reverse(self):
        """Flip the direction of the edge, accounting for curvatures"""

        self.start, self.end = self.end, self.start
        self.control_y *= -1

        return self
    
    def reflect_features(self):
        """Reflect edge fetures from one side of the edge to the other"""

        self.control_y *= -1

        return self

    def _subdivide(self, fractions: list, by_length=False):
        """Add intermediate vertices to an edge, 
            splitting it's parametrization according to fractions
            while preserving the overall shape
        
            NOTE: param subdiv == length subdiv for circle arcs
        """
        # NOTE: subdivide_param() is the same as subdivide_len()
        # So parent implementation is ok
        frac = [abs(f) for f in fractions]
        if not close_enough(fsum:=sum(frac), 1, 1e-4):
            raise RuntimeError(f'Edge Subdivision::Error::fraction is incorrect. The sum {fsum} is not 1')

        curve = self.as_curve()
        # Sub-curves
        covered_fr = 0
        subcurves = []
        for fr in fractions:
            subcurves.append(curve.cropped(covered_fr, covered_fr + fr))
            covered_fr += fr

        # Convert to CircleEdge objects
        subedges = EdgeSequence()
        for curve in subcurves:
            subedges.append(CircleEdge.from_svg_curve(curve))
        # Reference the first/last vertices correctly
        subedges[0].start = self.start
        subedges[-1].end = self.end

        return subedges

    # Special tools for circle representation
    def as_curve(self):
        """Represent as svgpath Arc"""

        radius, la, sweep = self.as_radius_flag()

        return svgpath.Arc(
            list_to_c(self.start),
            list_to_c([radius, radius]), 0, la, sweep,
            list_to_c(self.end)
        )
  
    def as_radius_flag(self):
        """Return circle representation as radius and arc flags"""

        return (self._rel_radius() * self._straight_len(), 
                self._is_large_arc(),
                self.control_y < 0)   # left/right orientation 
  
    def as_radius_angle(self):
        """Return circle representation as radius and an angle"""

        return (
            self._rel_radius() * self._straight_len(),
            self._arc_angle(),
            self.control_y < 0
        )

    def linearize(self):
        """Return a linear approximation of an edge using the same vertex objects
        
            # NOTE: Add extra vertex at an extremum of the edge
        """
        midpoint = self.midpoint()
        edges = [Edge(self.start, midpoint), Edge(midpoint, self.end)]

        return EdgeSequence(*edges)

    # NOTE: The following values are calculated at runtime to allow 
    # changes to control point after the edge definition
    def _rel_radius(self, abs_radius=None):
        """Eval relative radius (w.r.t. straight distance) from 3-point representation"""

        if abs_radius: 
            return abs_radius / self._straight_len()

        # Using the formula for radius of circumscribed circle
        # https://en.wikipedia.org/wiki/Circumscribed_circle#Other_properties

        # triangle sides, assuming the begginning and end of an edge are at (0, 0) and (1, 0)
        # accordingly
        a = 1
        b = norm([0.5, self.control_y])
        c = norm([0.5 - 1, self.control_y])
        p = (a + b + c) / 2  # semiperimeter

        rad = a * b * c / np.sqrt(p * (p - a) * (p - b) * (p - c)) / 4

        return rad

    def _arc_angle(self):
        """Eval arc angle from control point"""
        rel_rad = self._rel_radius()
        arc = 2 * np.arcsin(1 / rel_rad / 2)

        if self._is_large_arc():
            arc = 2 * np.pi - arc
        
        return arc
    
    def _is_large_arc(self):
        """Indicate if the arc sweeps the large or small angle"""
        return abs(self.control_y) > self._rel_radius()
    

    # Factories
    @staticmethod
    def from_points_angle(start, end, arc_angle, right=True):
        """Construct circle arc from two fixed points and an angle
        
            arc_angle: 
            
            NOTE: Might fail on angles close to 2pi
        """
        # Big or small arc
        if arc_angle > np.pi:
            arc_angle = 2*np.pi - arc_angle
            to_sum = True
        else: 
            to_sum = False

        radius = 1 / np.sin(arc_angle / 2) / 2
        h = 1 / np.tan(arc_angle / 2) / 2

        control_y = radius + h if to_sum else radius - h  # relative control point
        control_y *= -1 if right else 1

        return CircleEdge(start, end, cy=control_y)

    @staticmethod
    def from_points_radius(start, end, radius, large_arc=False, right=True):
        """Construct circle arc relative representation
            from two fixed points and an (absolute) radius
        """
        # Find circle center
        str_dist = norm(np.asarray(end) - np.asarray(start))
        center_r = np.sqrt(radius**2 - str_dist**2 / 4)

        # Find the absolute value of Y
        control_y = radius + center_r if large_arc else radius - center_r

        # Convert to relative
        control_y = control_y / str_dist

        # Flip sight according to "right" parameter
        control_y *= -1 if right else 1 

        return CircleEdge(start, end, cy=control_y)

    @staticmethod
    def from_svg_curve(seg:svgpath.Arc):
        """Create object from svgpath arc"""
        start, end = c_to_list(seg.start), c_to_list(seg.end)
        # NOTE: assuming circular arc (same radius in both directoins)
        radius = seg.radius.real

        return CircleEdge.from_points_radius(
            start, end, radius, seg.large_arc, seg.sweep
        )

    @staticmethod
    def from_three_points(start, end, point_on_arc):
        """Create a circle arc from 3 points (start, end and any point on an arc)
        
            NOTE: Control point specified in the same coord system as start and end
            NOTE: points should not be on the same line
        """

        nstart, nend, npoint_on_arc = np.asarray(start), np.asarray(end), np.asarray(point_on_arc)

        # https://stackoverflow.com/a/28910804
        # Using complex numbers to calculate the center & radius
        x, y, z = list_to_c([start, point_on_arc, end]) 
        w = z - x
        w /= y - x
        c = (x - y)*(w - abs(w)**2)/2j/w.imag - x
        # NOTE center = [c.real, c.imag]
        rad = abs(c + x)

        # Large/small arc
        mid_dist = norm(npoint_on_arc - ((nstart + nend) / 2))

        # Orientation
        angle = vector_angle(npoint_on_arc - nstart, nend - nstart)  # +/-

        return CircleEdge.from_points_radius(
            start, end, radius=rad, 
            large_arc=mid_dist > rad, right=angle > 0) 

    # Finally
    def assembly(self):
        """Returns the dict-based representation of edges, 
            compatible with core -> BasePattern JSON (dict) 
        """

        rad, large_arc, right = self.as_radius_flag()
        return (
            [self.start, self.end], 
            {
                "endpoints": [0, 1], 
                "curvature": {
                    "type": 'circle',
                    "params": [rad, int(large_arc), int(right)]
                }
            })


class CurveEdge(Edge):
    """Curvy edge as Besier curve / B-spline"""

    def __init__(self, start=[0, 0], end=[0, 0], control_points=[], relative=True) -> None:
        """Define a Besier curve edge
            * start, end: from/to vertcies that the edge connectes
            * control_points: coordinated of Bezier control points.
                Specification of One control point creates the Quadratic Bezier, 
                Specification of 2 control points creates Cubic Bezier. 
                Other degrees are not supported.

            * relative: specify whether the control point coordinated are given 
            relative to the edge length (True) or in 2D coordinate system of a panel (False)

        """
        super().__init__(start, end)

        self.control_points = control_points

        if len(self.control_points) > 2:
            raise NotImplementedError(f'{self.__class__.__name__}::Error::Up to 2 control points (cubic Bezier) are supported')

        # Storing control points as relative since it preserves overall curve shape during
        # edge extention/contration
        if not relative:
            self.control_points = [self._abs_to_rel_2d(c).tolist() for c in self.control_points]

    def length(self):
        """Length of Bezier curve edge"""
        curve = self.as_curve()
        
        return curve.length()

    def __str__(self) -> str:

        points = [self.start] + self.control_points

        str = [f'[{p[0]:.2f}, {p[1]:.2f}]->' for p in points]
        str += [f'[{self.end[0]:.2f}, {self.end[1]:.2f}]']

        return 'Curve:' + ''.join(str)
    
    def midpoint(self):
        """Center of the edge"""
        curve = self.as_curve()

        t_mid = curve.ilength(curve.length()/2)
        return c_to_list(curve.point(t_mid))
    
    def _subdivide(self, fractions: list, by_length=False):
        """Add intermediate vertices to an edge, 
            splitting its curve parametrization or overall length according to 
            fractions while preserving the overall shape
        """
        curve = self.as_curve()

        # Sub-curves
        covered_fr, prev_t = 0, 0
        clen = curve.length()
        subcurves = []
        for fr in fractions:
            covered_fr += fr
            if by_length:
                next_t = curve.ilength(clen * covered_fr)
                subcurves.append(curve.cropped(prev_t, next_t))
                prev_t = next_t
            else:
                subcurves.append(curve.cropped(covered_fr - fr, covered_fr))

        # Convert to CurveEdge objects
        subedges = EdgeSequence()
        for curve in subcurves:
            subedges.append(CurveEdge.from_svg_curve(curve))
        # Reference the first/last vertices correctly
        subedges[0].start = self.start
        subedges[-1].end = self.end

        return subedges

    # Actions
    def reverse(self):
        """Flip the direction of the edge, accounting for curvatures"""

        self.start, self.end = self.end, self.start

        # change order of control points
        if len(self.control_points) == 2:
            self.control_points[0], self.control_points[1] = self.control_points[1], self.control_points[0]

        # Update coordinates
        for p in self.control_points:
            p[0], p[1] = 1 - p[0], -p[1]

        return self
    
    def reflect_features(self):
        """Reflect edge fetures from one side of the edge to the other"""

        for p in self.control_points:
            p[1] = -p[1]

        return self
    
    def as_curve(self, absolute=True):
        """As svgpath curve object

            Converting on the fly as exact vertex location might have been updated since
            the creation of the edge
        """
        # Get the nodes correcly
        if absolute:
            cp = [self._rel_to_abs_2d(c) for c in self.control_points]
            nodes = np.vstack((self.start, cp, self.end))
        else:
            cp = self.control_points
            nodes = np.vstack(([0, 0], cp, [1, 0]))
        
        params = nodes[:, 0] + 1j*nodes[:, 1]

        return svgpath.QuadraticBezier(*params) if len(cp) < 2 else svgpath.CubicBezier(*params)

    def linearize(self):
        """Return a linear approximation of an edge using the same vertex objects
        
            # NOTE: Add extra vertex at an extremum of the edge
        """
        extreme_points = self._extreme_points()

        if len(extreme_points):
            seq = EdgeSequence(Edge(self.start, extreme_points[0]))
            for i in range(1, len(extreme_points)):
                seq.append(Edge(seq[-1].end, extreme_points[i]))
            seq.append(Edge(seq[-1].end, self.end))

            return seq
        else:
            return Edge(self.start, self.end)

    def _extreme_points(self):
        """Return extreme points (on Y) of the current edge
            NOTE: this does NOT include the border vertices of an edge
        """

        # Variation of https://github.com/mathandy/svgpathtools/blob/5c73056420386753890712170da602493aad1860/svgpathtools/bezier.py#L197
        curve = self.as_curve(absolute=False)   # relative coords to find real extremizers
        poly = svgpath.bezier2polynomial(curve, return_poly1d=True)
        y = svgpath.imag(poly)
        dy = y.deriv()
        y_extremizers = svgpath.polyroots(dy, realroots=True,
                                          condition=lambda r: 0 < r < 1)

        extreme_points = np.array(
            [self._rel_to_abs_2d(c_to_list(curve.point(t))) for t in y_extremizers]
        )

        return extreme_points

    @staticmethod
    def from_svg_curve(seg):
        """Create CurveEdge object from svgpath bezier objects"""

        start, end = c_to_list(seg.start), c_to_list(seg.end)
        if isinstance(seg, svgpath.QuadraticBezier):
            cp = [c_to_list(seg.control)]
        elif isinstance(seg, svgpath.CubicBezier):
            cp = [c_to_list(seg.control1), c_to_list(seg.control2)]
        else:
            raise NotImplementedError(f'CurveEdge::Error::Incorrect curve type supplied {seg.type}')

        return CurveEdge(start, end, cp, relative=False)

    # Assembly into serializable object
    def assembly(self):
        """Returns the dict-based representation of edges, 
            compatible with core -> BasePattern JSON (dict) 
        """

        return (
            [self.start, self.end], 
            {
                "endpoints": [0, 1], 
                "curvature": {  
                    "type": 'quadratic' if len(self.control_points) == 1 else 'cubic',
                    "params": self.control_points
                }
            })

    
class EdgeSequence():
    """Represents a sequence of (possibly chained) edges (e.g. every next edge starts from the same vertex that the previous edge ends with
        and allows building some typical edge sequences
    """
    def __init__(self, *args) -> None:
        self.edges = []
        for arg in args:
            self.append(arg)

    # ANCHOR Properties
    def __getitem__(self, i):
        if isinstance(i, slice):
            # return an EdgeSequence object for slices
            e_slice = self.edges[i]
            return EdgeSequence(e_slice)
        else:
            return self.edges[i]

    def index(self, elem):
        # Find the same object (by reference) 
        # list.index() is doing something different..
        # https://stackoverflow.com/a/47057419
        return next(i for i, e in enumerate(self.edges) if elem is e)

    def __len__(self):
        """Number of edges in the sequence"""
        return len(self.edges)

    def __contains__(self, item):
        # check presence by comparing references
        return any([item is e for e in self.edges])

    def __str__(self) -> str:
        return 'EdgeSeq: ' + str(self.edges)
    
    def __repr__(self) -> str:
        return self.__str__()

    def length(self):
        """Total length of edges"""
        return sum([e.length() for e in self.edges])

    def isLoop(self):
        return self.edges[0].start is self.edges[-1].end and len(self) > 1

    def isChained(self):
        """Does the sequence of edges represent correct chain?"""
        if len(self) < 2:
            return False

        for i in range(1, len(self.edges)):
            if self.edges[i].start is not self.edges[i-1].end:
                if VERBOSE:
                    # This should be helpful to catch bugs
                    print(f'{self.__class__.__name__}::Warning!::Edge sequence is not properly chained')
                return False
        return True

    def fractions(self) -> list:
        """Fractions of the lengths of each edge in sequence w.r.t. 
            the whole sequence
        """
        total_len = sum([e.length() for e in self.edges])

        return [e.length() / total_len for e in self.edges]

    def lengths(self) -> list:
        """Lengths of individual edges in the sequence"""
        return [e.length() for e in self.edges]

    def verts(self):
        """Return all vertex objects"""
        verts = [self.edges[0].start]
        for e in self.edges:
            if e.start is not verts[-1]:  # avoid adding the vertices of chained edges twice
                verts.append(e.start)
            verts.append(e.end)
        if verts[0] is verts[-1]:  # don't double count the loop origin
            verts.pop(-1)
        return verts

    def shortcut(self):
        """Opening of an edge sequence as a vector
        
            # NOTE May not reflect true shortcut if the egdes were flipped but the order remained
        """
        return np.array([self[0].start, self[-1].end]) 

    # ANCHOR Modifiers
    # All modifiers return self object to allow chaining
    # Wrappers around python's list
    def append(self, item):
        if isinstance(item, Edge):
            self.edges.append(item)
        elif isinstance(item, list):  # List of edge / EdgeSeq objects
            for e in item:
                self.append(e)
        elif isinstance(item, EdgeSequence):
            self.edges += item.edges
        else:
            raise ValueError(f'{self.__class__.__name__}::Error::Trying to add object of incompatible type {type(item)}')
        return self

    def insert(self, i, item):
        if isinstance(item, Edge):
            self.edges.insert(i, item)
        elif isinstance(item, list) or isinstance(item, EdgeSequence):
            for j in range(len(item)):
                self.edges.insert(i + j, item[j])
        else:
            raise NotImplementedError(f'{self.__class__.__name__}::Error::incerting object of {type(item)} not suported (yet)')
        return self
    
    def pop(self, i):
        if isinstance(i, Edge):
            i = self.index(i)
        self.edges.pop(i)
        return self

    def substitute(self, orig, new):
        """Remove orign item from the list and place seq into it's place
            orig can be either an id of an item to remove 
            or an instance of Edge that exists in the current sequence
        """
        if isinstance(orig, Edge):
            orig = self.index(orig)
        if orig < 0: 
            orig = len(self) + orig 
        self.pop(orig)
        self.insert(orig, new)
        return self

    def reverse(self):
        """Reverse edge sequence in-place"""
        self.edges.reverse()
        for edge in self.edges:
            edge.reverse()
        return self

    # EdgeSequence-specific
    def translate_by(self, shift):
        """Translate the edge seq vertices s.t. the first vertex is at new_origin
        """
        for v in self.verts():
            v[0] += shift[0]
            v[1] += shift[1]
        return self

    def snap_to(self, new_origin=[0, 0]):
        """Translate the edge seq vertices s.t. the first vertex is at new_origin
        """
        start = copy(self[0].start)
        shift = [new_origin[0] - start[0], new_origin[1] - start[1]]
        self.translate_by(shift)

        return self

    def close_loop(self):
        """if edge loop is not closed, add and edge to close it"""
        self.isChained()  # print warning if smth is wrong
        if not self.isLoop():
            self.append(Edge(self[-1].end, self[0].start))
        return self

    def rotate(self, angle):
        """Rotate edge sequence by angle in place, using first point as a reference

        Parameters: 
            angle -- desired rotation angle in radians (!)
        """
        curr_start = copy(self[0].start)
        
        # set the start point to zero
        self.snap_to([0, 0])
        rot = R2D(angle)

        for v in self.verts():
            v[:] = np.matmul(rot, v)
        
        # recover the original location
        self.snap_to(curr_start)

        return self

    def extend(self, factor):
        """Extend or shrink the edges along the line from start of the first edge to the 
        end of the last edge in sequence
        """
        # Need to take the target line from the chained order
        if not self.isChained():  
            chained_edges = self.chained_order()
            chained_edges.isChained()
            if chained_edges.isLoop():
                print(f'{self.__class__.__name__}::Warning::Extending looped edge sequences is not available')
                return self
        else: 
            chained_edges = self
        
        target_line = np.array(chained_edges[-1].end) - np.array(chained_edges[0].start)
        target_line = target_line / norm(target_line)

        # gather vertices
        verts_coords = self.verts()
        nverts_coords = np.array(verts_coords)
        
        # adjust their position based on projection to the target line
        verts_projection = np.empty(nverts_coords.shape)
        fixed = nverts_coords[0]
        for i in range(nverts_coords.shape[0]):
            verts_projection[i] = (nverts_coords[i] - fixed).dot(target_line) * target_line

        new_verts = verts_coords - (1 - factor) * verts_projection

        # Update vertex objects
        for i in range(len(verts_coords)):
            verts_coords[i][:] = new_verts[i]

        return self

    def reflect(self, v0, v1):
        """Reflect 2D points w.r.t. 1D line defined by two points"""
        v0, v1 = np.asarray(v0), np.asarray(v1)
        vec = np.asarray(v1) - np.asarray(v0)
        vec = vec / norm(vec)  # normalize

        # https://demonstrations.wolfram.com/ReflectionMatrixIn2D/#more
        Ref = np.array([
            [ 1 - 2 * vec[1]**2,  2*vec[0]*vec[1]],
            [ 2*vec[0]*vec[1],    - 1 + 2 * vec[1]**2 ]
            ])
        
        # translate -> reflect -> translate back
        for v in self.verts():
            v[:] = np.matmul(Ref, np.asarray(v) - v0) + v0

        # Reflect edge features (curvatures, etc.)
        for e in self.edges:
            e.reflect_features()

        return self

    # ANCHOR New sequences & versions
    def copy(self):
        """Create a copy of a current edge sequence preserving the chaining property of edge sequences"""
        new_seq = deepcopy(self)

        # deepcopy recreates the vertex objects on both sides of the edges
        # in chaned edges those vertex objects are supposed to be shared
        # by neighbor edges

        for i in range(1, len(new_seq)):
            if self[i].start is self[i-1].end:
                new_seq[i].start = new_seq[i-1].end
            
        if self.isLoop():
            new_seq[-1].end = new_seq[0].start

        return new_seq

    def chained_order(self):
        """ Attempt to restore a chain in the EdgeSequence
            The chained edge sequence may loose it's property if the edges were reveresed externally. 
            This routine created a copy of the currect sequence with aligned the order of edges,

            It might be useful for various calculations
        
        """
        chained = self.copy()
        
        for i in range(len(chained)):
            # Assuming the previous one is already sorted
            if i > 0 and chained[i].end is chained[i-1].end:
                chained[i].reverse()
            # Not connected to the previous one
            elif (i + 1 < len(chained)
                    and (chained[i].start is chained[i+1].start or chained[i].start is chained[i+1].end)):
                chained[i].reverse()
            # not connected to anything or connected properly -- leave as is
        
        return chained


"""
    Module for basic operations on patterns
"""
# Basic
import copy
from datetime import datetime
import errno
import json
import numpy as np
import os
import random

# My
from pattern import rotation as rotation_tools

standard_filenames = [
    'specification',  # e.g. used by dataset generation
    'template', 
    'prediction'
]

pattern_spec_template = {
    'pattern': {
        'panels': {},
        'stitches': []
    },
    'parameters': {},
    'parameter_order': [],
    'properties': {  # these are to be ensured when pattern content is updated directly
        'curvature_coords': 'relative', 
        'normalize_panel_translation': False, 
        'normalized_edge_loops': True,  # will trigger edge loop normalization on reload
        'units_in_meter': 100  # cm
    }
}

panel_spec_template = {
    'translation': [ 0, 0, 0 ],
    'rotation': [ 0, 0, 0 ],
    'vertices': [],
    'edges': []
}


# ------------ Patterns --------
class BasicPattern(object):
    """Loading & serializing of a pattern specification in custom JSON format.
        Input:
            * Pattern template in custom JSON format
        Output representations: 
            * Pattern instance in custom JSON format 
                * In the current state
        
        Not implemented: 
            * Convertion to NN-friendly format
            * Support for patterns with darts
    """

    # ------------ Interface -------------

    def __init__(self, pattern_file=None):
        
        self.spec_file = pattern_file
        
        if pattern_file is not None: # load pattern from file
            self.path = os.path.dirname(pattern_file)
            self.name = BasicPattern.name_from_path(pattern_file)
            self.reloadJSON()
        else: # create empty pattern
            self.path = None
            self.name = self.__class__.__name__
            self.spec = copy.deepcopy(pattern_spec_template)
            self.pattern = self.spec['pattern']
            self.properties = self.spec['properties']  # mandatory part

    def reloadJSON(self):
        """(Re)loads pattern info from spec file. 
        Useful when spec is updated from outside"""
        if self.spec_file is None:
            print('BasicPattern::Warning::{}::Pattern is not connected to any file. Reloadig from file request ignored.'.format(
                self.name
            ))
            return

        with open(self.spec_file, 'r') as f_json:
            self.spec = json.load(f_json)
        self.pattern = self.spec['pattern']
        self.properties = self.spec['properties']  # mandatory part

        # template normalization - panel translations and curvature to relative coords
        self._normalize_template()

    def serialize(self, path, to_subfolder=True, tag='', empty_ok=False):

        if not empty_ok and len(self.panel_order()) == 0:
            raise RuntimeError(f'{self.__class__.__name__}::ERROR::Asked to save an empty pattern')

        # log context
        if to_subfolder:
            log_dir = os.path.join(path, self.name + '_' + tag)  # NOTE Added change
            try:
                os.makedirs(log_dir)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise
            spec_file = os.path.join(log_dir, tag + 'specification.json')
        else:
            log_dir = path
            spec_file = os.path.join(path, (self.name + tag + '_specification.json'))

        # Save specification
        with open(spec_file, 'w') as f_json:
            json.dump(self.spec, f_json, indent=2)
        
        return log_dir

    @staticmethod
    def name_from_path(pattern_file):
        name = os.path.splitext(os.path.basename(pattern_file))[0]
        if name in standard_filenames:  # use name of directory instead
            path = os.path.dirname(pattern_file)
            name = os.path.basename(os.path.normpath(path))
        return name

    # --------- Info ------------------------
    def panel_order(self, force_update=False):
        """
            Return current agreed-upon order of panels
            * if not defined in the pattern or if  'force_update' is enabled, re-evaluate it based on curent panel translation and save
        """
        if 'panel_order' not in self.pattern or force_update:
            self.pattern['panel_order'] = self.define_panel_order()
        return self.pattern['panel_order']

    def define_panel_order(self, name_list=None, location_dict=None, dim=0, tolerance=10):
        """ (Recursive) Ordering of the panels based on their 3D translation values.
            * Using cm as units for tolerance (when the two coordinates are considered equal)
            * Sorting by all dims as keys X -> Y -> Z (left-right (looking from Z) then down-up then back-front)
            * based on the fuzzysort suggestion here https://stackoverflow.com/a/24024801/11206726"""

        if name_list is None:  # start from beginning
            name_list = self.pattern['panels'].keys() 
        if not name_list:
            return []
        if location_dict is None:  # obtain location for all panels to use in sorting further
            location_dict = {}
            for name in name_list:
                location_dict[name], _ = self._panel_universal_transtation(name)

        # consider only translations of the requested panel names
        reference = [location_dict[panel_n][dim] for panel_n in name_list]
        sorted_couple = sorted(zip(reference, name_list))  # sorts according to the first list
        sorted_reference, sorted_names = zip(*sorted_couple)
        sorted_names = list(sorted_names)

        if (dim + 1) < 3:  # 3D is max
            # re-sort values by next dimention if they have similar values in current dimention
            fuzzy_start, fuzzy_end = 0, 0  # init both in case we start from 1 panel to sort
            for fuzzy_end in range(1, len(sorted_reference)):
                if sorted_reference[fuzzy_end] - sorted_reference[fuzzy_start] >= tolerance:
                    # the range of similar values is completed
                    if fuzzy_end - fuzzy_start > 1:
                        sorted_names[fuzzy_start:fuzzy_end] = self.define_panel_order(
                            sorted_names[fuzzy_start:fuzzy_end], location_dict, dim + 1, tolerance)
                    fuzzy_start = fuzzy_end  # start counting similar values anew

            # take care of the tail
            if fuzzy_start != fuzzy_end:
                sorted_names[fuzzy_start:] = self.define_panel_order(
                    sorted_names[fuzzy_start:], location_dict, dim + 1, tolerance)

        return sorted_names

    # -- sub-utils --
    def _edge_as_vector(self, vertices, edge_dict):
        """Represent edge as vector of fixed length: 
            * First 2 elements: Vector endpoint. 
                Original edge endvertex positions can be restored if edge vector is added to the start point,
                which in turn could be obtained from previous edges in the panel loop
            * Next 2 elements: Curvature values 
                Given in relative coordinates. With zeros if edge is not curved 

        """
        edge_verts = vertices[edge_dict['endpoints']]
        edge_vector = edge_verts[1] - edge_verts[0]
        curvature = np.array(edge_dict['curvature']) if 'curvature' in edge_dict else [0, 0]

        return np.concatenate([edge_vector, curvature])

    @staticmethod
    def _point_in_3D(local_coord, rotation, translation):
        """Apply 3D transformation to the point given in 2D local coordinated, e.g. on the panel
        * rotation is expected to be given in 'xyz' Euler anges (as in Autodesk Maya) or as 3x3 matrix"""

        # 2D->3D local
        local_coord = np.append(local_coord, 0)

        # Rotate
        rotation = np.array(rotation)
        if rotation.size == 3:  # transform Euler angles to matrix
            rotation = rotation_tools.euler_xyz_to_R(rotation)
            # otherwise we already have the matrix
        elif rotation.size != 9:
            raise ValueError('BasicPattern::Error::You need to provide Euler angles or Rotation matrix for _point_in_3D(..)')
        rotated_point = rotation.dot(local_coord)

        # translate
        return rotated_point + translation

    def _panel_universal_transtation(self, panel_name):
        """Return a universal 3D translation of the panel (e.g. to be used in judging the panel order).
            Universal translation it defined as world 3D location of mid-point of the top (in 3D) of the panel (2D) bounding box.
            * Assumptions: 
                * In most cases, top-mid-point of a panel corresponds to body landmarks (e.g. neck, middle of an arm, waist) 
                and thus is mostly stable across garment designs.
                * 3D location of a panel is placing this panel around the body in T-pose
            * Function result is independent from the current choice of the local coordinate system of the panel
        """
        panel = self.pattern['panels'][panel_name]
        vertices = np.array(panel['vertices'])

        # out of 2D bounding box sides' midpoints choose the one that is highest in 3D
        top_right = vertices.max(axis=0)
        low_left = vertices.min(axis=0)
        mid_x = (top_right[0] + low_left[0]) / 2
        mid_y = (top_right[1] + low_left[1]) / 2
        mid_points_2D = [
            [mid_x, top_right[1]], 
            [mid_x, low_left[1]],
            [top_right[0], mid_y],
            [low_left[0], mid_y]
        ]
        rot_matrix = rotation_tools.euler_xyz_to_R(panel['rotation'])  # calculate once for all points 
        mid_points_3D = np.vstack(tuple(
            [self._point_in_3D(coords, rot_matrix, panel['translation']) for coords in mid_points_2D]
        ))
        top_mid_point = mid_points_3D[:, 1].argmax()

        return mid_points_3D[top_mid_point], np.array(mid_points_2D[top_mid_point])

    # --------- Pattern operations (changes inner dicts) ----------
    def _normalize_template(self):
        """
        Updated template definition for convenient processing:
            * Converts curvature coordinates to realitive ones (in edge frame) -- for easy length scaling
            * snaps each panel center to (0, 0) if requested in props
            * scales everything to cm
        """
        if self.properties['curvature_coords'] == 'absolute':
            for panel in self.pattern['panels']:
                # convert curvature 
                vertices = self.pattern['panels'][panel]['vertices']
                edges = self.pattern['panels'][panel]['edges']
                for edge in edges:
                    if 'curvature' in edge:
                        edge['curvature'] = self._control_to_relative_coord(
                            vertices[edge['endpoints'][0]], 
                            vertices[edge['endpoints'][1]], 
                            edge['curvature']
                        )
            # now we have new property
            self.properties['curvature_coords'] = 'relative'
        
        if 'units_in_meter' in self.properties:
            if self.properties['units_in_meter'] != 100:
                for panel in self.pattern['panels']:
                    self._normalize_panel_scaling(panel, self.properties['units_in_meter'])
                # now we have cm
                self.properties['original_units_in_meter'] = self.properties['units_in_meter']
                self.properties['units_in_meter'] = 100
                print('Warning: pattern units converted to cm')
        else:
            print('Warning: units not specified in the pattern. Scaling normalization was not applied')

        # after curvature is converted!!
        # Only if requested
        if ('normalize_panel_translation' in self.properties 
                and self.properties['normalize_panel_translation']):
            print('Normalizing translation!')
            self.properties['normalize_panel_translation'] = False  # one-time use property. Preverts rotation issues on future reads
            for panel in self.pattern['panels']:
                # put origin in the middle of the panel-- 
                offset = self._normalize_panel_translation(panel)
                # udpate translation vector
                original = self.pattern['panels'][panel]['translation'] 
                self.pattern['panels'][panel]['translation'] = [
                    original[0] + offset[0], 
                    original[1] + offset[1], 
                    original[2], 
                ]

        # Recalculate origins and traversal order of panel edge loops if not normalized already
        if ('normalized_edge_loops' not in self.properties
                or not self.properties['normalized_edge_loops']):
            print('{}::Warning::normalizing the order and origin choice for edge loops in panels'.format(self.__class__.__name__))
            self.properties['normalized_edge_loops'] = True
            for panel in self.pattern['panels']:
                self._normalize_edge_loop(panel)
        
        # Recalculate panel order if not given already
        self.panel_order()

    def _normalize_panel_translation(self, panel_name):
        """ Convert panel vertices to local coordinates: 
            Shifts all panel vertices s.t. origin is at the center of the panel
        """
        panel = self.pattern['panels'][panel_name]
        vertices = np.asarray(panel['vertices'])
        offset = np.mean(vertices, axis=0)
        vertices = vertices - offset

        panel['vertices'] = vertices.tolist()

        return offset
    
    def _normalize_panel_scaling(self, panel_name, units_in_meter):
        """Convert all panel info to cm. I assume that curvature is alredy converted to relative coords -- scaling does not need update"""
        scaling = 100 / units_in_meter
        # vertices
        vertices = np.array(self.pattern['panels'][panel_name]['vertices'])
        vertices = scaling * vertices
        self.pattern['panels'][panel_name]['vertices'] = vertices.tolist()

        # translation
        translation = self.pattern['panels'][panel_name]['translation']
        self.pattern['panels'][panel_name]['translation'] = [scaling * coord for coord in translation]

    def _normalize_edge_loop(self, panel_name):
        """
            * Re-order edges s.t. the edge loop starts from low-left vertex
            * Make the edge loop follow counter-clockwise direction (uniform traversal)
        """
        panel = self.pattern['panels'][panel_name]
        vertices = np.array(panel['vertices'])

        # Loop Origin
        loop_origin_id = self._vert_at_left_corner(vertices)
        print('{}:{}: Origin: {} -> {}'.format(
            self.name, panel_name, panel['edges'][0]['endpoints'][0], loop_origin_id))

        rotated_edges, rotated_edge_ids = self._rotate_edges(
            panel['edges'], list(range(len(panel['edges']))), loop_origin_id)
        panel['edges'] = rotated_edges

        # Panel flip for uniform edge loop order (and normal direction)
        first_edge = self._edge_as_vector(vertices, rotated_edges[0])[:2]
        last_edge = self._edge_as_vector(vertices, rotated_edges[-1])[:2]
        flipped = False
        # due to the choice of origin (at the corner), first & last edge cross-product will reliably show panel normal direction 
        if np.cross(first_edge, last_edge) > 0:  # should be negative -- counterclockwise
            print('{}::{}::panel <{}> flipped'.format(
                self.__class__.__name__, self.name, panel_name
            ))
            flipped = True

            # Vertices
            vertices[:, 0] = - vertices[:, 0]  # flip by X coordinate -- we'll rotate around Y
            panel['vertices'] = vertices.tolist()

            # Edges
            # new loop origin after update
            loop_origin_id = self._vert_at_left_corner(vertices)
            print('{}:{}: Origin: {} -> {}'.format(
                self.name, panel_name, panel['edges'][0]['endpoints'][0], loop_origin_id))

            rotated_edges, rotated_edge_ids = self._rotate_edges(rotated_edges, rotated_edge_ids, loop_origin_id)
            panel['edges'] = rotated_edges
            # update the curvatures in edges as they changed left\right symmetry in 3D
            for edge_id in range(len(rotated_edges)):
                if 'curvature' in panel['edges'][edge_id]:
                    curvature = panel['edges'][edge_id]['curvature']
                    # YES!! Only one of the curvature coordinates need update at this point
                    panel['edges'][edge_id]['curvature'][1] = -curvature[1]

            # Panel translation and rotation -- local coord frame changed!
            panel['translation'][0] -= 2 * panel['translation'][0] 

            panel_R = rotation_tools.euler_xyz_to_R(panel['rotation'])
            flip_R = np.eye(3)
            flip_R[0, 0] = flip_R[2, 2] = -1  # by 180 around Y

            panel['rotation'] = rotation_tools.R_to_euler(panel_R * flip_R)

        # Stitches -- update the edge references according to the new ids
        for stitch_id in range(len(self.pattern['stitches'])):
            for side_id in [0, 1]:
                if self.pattern['stitches'][stitch_id][side_id]['panel'] == panel_name:
                    old_edge_id = self.pattern['stitches'][stitch_id][side_id]['edge']
                    self.pattern['stitches'][stitch_id][side_id]['edge'] = rotated_edge_ids[old_edge_id]

        return rotated_edge_ids, flipped

    # -- sub-utils --
    def _control_to_abs_coord(self, start, end, control_scale):
        """
        Derives absolute coordinates of Bezier control point given as an offset
        """
        edge = end - start
        edge_perp = np.array([-edge[1], edge[0]])

        control_start = start + control_scale[0] * edge
        control_point = control_start + control_scale[1] * edge_perp

        return control_point 
    
    def _control_to_relative_coord(self, start, end, control_point):
        """
        Derives relative (local) coordinates of Bezier control point given as 
        a absolute (world) coordinates
        """
        start, end, control_point = np.array(start), np.array(end), \
            np.array(control_point)

        control_scale = [None, None]
        edge_vec = end - start
        edge_len = np.linalg.norm(edge_vec)
        control_vec = control_point - start
        
        # X
        # project control_vec on edge_vec by dot product properties
        control_projected_len = edge_vec.dot(control_vec) / edge_len 
        control_scale[0] = control_projected_len / edge_len
        # Y
        control_projected = edge_vec * control_scale[0]
        vert_comp = control_vec - control_projected  
        control_scale[1] = np.linalg.norm(vert_comp) / edge_len
        # Distinguish left&right curvature
        control_scale[1] *= np.sign(np.cross(control_point, edge_vec))

        return control_scale 

    def _edge_length(self, panel, edge):
        panel = self.pattern['panels'][panel]
        v_id_start, v_id_end = tuple(panel['edges'][edge]['endpoints'])
        v_start, v_end = np.array(panel['vertices'][v_id_start]), \
            np.array(panel['vertices'][v_id_end])
        
        return np.linalg.norm(v_end - v_start)

    @staticmethod
    def _vert_at_left_corner(vertices):
        """
            Find, which vertex is in the left corner
            * Determenistic process
        """
        left_corner = np.min(vertices, axis=0)
        vertices = vertices - left_corner
        
        # choose the one closest to zero (=low-left corner) as new origin
        verts_norms = np.linalg.norm(vertices, axis=1)  # numpy 1.9+
        origin_id = np.argmin(verts_norms)

        return origin_id

    @staticmethod
    def _rotate_edges(edges, edge_ids, new_origin_id):
        """
            Rotate provided list of edges s.t. the first edge starts from vertex with id = new_origin_id
            Map old edge_ids to new ones accordingly
            * edges expects list of edges structures
        """

        first_edge_orig_id = [idx for idx, edge in enumerate(edges) if edge['endpoints'][0] == new_origin_id]

        first_edge_orig_id = first_edge_orig_id[0]
        rotated_edges = edges[first_edge_orig_id:] + edges[:first_edge_orig_id]

        # map from old ids to new ids
        rotated_edge_ids = edge_ids[(len(rotated_edges) - first_edge_orig_id):] + edge_ids[:(len(rotated_edges) - first_edge_orig_id)]

        return rotated_edges, rotated_edge_ids

    def _restore(self, backup_copy):
        """Restores spec structure from given backup copy 
            Makes a full copy of backup to avoid accidential corruption of backup
        """
        self.spec = copy.deepcopy(backup_copy)
        self.pattern = self.spec['pattern']
        self.properties = self.spec['properties']  # mandatory part

    # -------- Checks ------------
    def is_self_intersecting(self):
        """returns True if any of the pattern panels are self-intersecting"""
        return any(map(self._is_panel_self_intersecting, self.pattern['panels']))

    def _is_panel_self_intersecting(self, panel_name):
        """Checks whatever a given panel contains intersecting edges
        """
        panel = self.pattern['panels'][panel_name]
        vertices = np.array(panel['vertices'])

        # construct edge list in coordinates
        edge_list = []
        for edge in panel['edges']:
            edge_ids = edge['endpoints']
            edge_coords = vertices[edge_ids]
            if 'curvature' in edge and isinstance(edge['curvature'], list):
                # NOTE: Legacy curvature representation
                curv_abs = self._control_to_abs_coord(edge_coords[0], edge_coords[1], edge['curvature'])
                # view curvy edge as two segments
                # NOTE this aproximation might lead to False positives in intersection tests
                edge_list.append([edge_coords[0], curv_abs])
                edge_list.append([curv_abs, edge_coords[1]])
            else:
                edge_list.append(edge_coords.tolist())

        # simple pairwise checks of edges
        # Follows discussion in  https://math.stackexchange.com/questions/80798/detecting-polygon-self-intersection 
        for i1 in range(0, len(edge_list)):
            for i2 in range(i1 + 1, len(edge_list)):
                if self._is_segm_intersecting(edge_list[i1], edge_list[i2]):
                    return True
        
        return False          
        
    def _is_segm_intersecting(self, segment1, segment2):
        """Checks wheter two segments intersect 
            in the points interior to both segments"""
        # https://algs4.cs.princeton.edu/91primitives/
        def ccw(start, end, point):
            """A test whether three points form counterclockwize angle (>0) 
            Returns (<0) if they form clockwize angle
            0 if collinear"""
            return (end[0] - start[0]) * (point[1] - start[1]) - (point[0] - start[0]) * (end[1] - start[1])

        if (ccw(segment1[0], segment1[1], segment2[0]) * ccw(segment1[0], segment1[1], segment2[1]) >= 0
                or ccw(segment2[0], segment2[1], segment1[0]) * ccw(segment2[0], segment2[1], segment1[1]) >= 0):
            return False
        return True


class ParametrizedPattern(BasicPattern):
    """
        Extention to BasicPattern that can work with parametrized patterns
        Update pattern with new parameter values & randomize those parameters
    """
    def __init__(self, pattern_file=None):
        super(ParametrizedPattern, self).__init__(pattern_file)
        self.parameters = self.spec['parameters']

        self.parameter_defaults = {
            'length': 1,
            'additive_length': 0,
            'curve': 1
        }
        self.constraint_types = [
            'length_equality'
        ]

    def param_values_list(self):
        """Returns current values of all parameters as a list in the pattern defined parameter order"""
        value_list = []
        for parameter in self.spec['parameter_order']:
            value = self.parameters[parameter]['value']
            if isinstance(value, list):
                value_list += value
            else:
                value_list.append(value)
        return value_list

    def apply_param_list(self, values):
        """Apply given parameters supplied as a list of param_values_list() form"""

        self._restore_template(params_to_default=False)

        # set new values
        value_count = 0
        for parameter in self.spec['parameter_order']:
            last_value = self.parameters[parameter]['value']
            if isinstance(last_value, list):
                self.parameters[parameter]['value'] = [values[value_count + i] for i in range(len(last_value))]
                value_count += len(last_value)
            else:
                self.parameters[parameter]['value'] = values[value_count]
                value_count += 1
        
        self._update_pattern_by_param_values()

    def reloadJSON(self):
        """(Re)loads pattern info from spec file. 
        Useful when spec is updated from outside"""
        super(ParametrizedPattern, self).reloadJSON()

        self.parameters = self.spec['parameters']
        self._normalize_param_scaling()

    def _restore(self, backup_copy):
        """Restores spec structure from given backup copy 
            Makes a full copy of backup to avoid accidential corruption of backup
        """
        super(ParametrizedPattern, self)._restore(backup_copy)
        self.parameters = self.spec['parameters']
    
    # ---------- Parameters operations --------

    def _normalize_param_scaling(self):
        """Convert additive parameters to cm units"""

        if 'original_units_in_meter' in self.properties:    # pattern was scaled
            scaling = 100 / self.properties['original_units_in_meter']
            for parameter in self.parameters:
                if self.parameters[parameter]['type'] == 'additive_length': 
                    self.parameters[parameter]['value'] = scaling * self.parameters[parameter]['value']
                    self.parameters[parameter]['range'] = [
                        scaling * elem for elem in self.parameters[parameter]['range']
                    ]

            # now we have cm everywhere -- no need to keep units info
            self.properties.pop('original_units_in_meter', None)

            print('Warning: Parameter units were converted to cm')

    def _normalize_edge_loop(self, panel_name):
        """Update the edge loops and edge ids references in parameters & constraints after change"""
        rotated_edge_ids, flipped = super(ParametrizedPattern, self)._normalize_edge_loop(panel_name)

        # Parameters
        for parameter_name in self.spec['parameters']:
            self._influence_after_edge_loop_update(
                self.spec['parameters'][parameter_name]['influence'], 
                panel_name, rotated_edge_ids)

        # Constraints
        if 'constraints' in self.spec:
            for constraint_name in self.spec['constraints']:
                self._influence_after_edge_loop_update(
                    self.spec['constraints'][constraint_name]['influence'], 
                    panel_name, rotated_edge_ids)

    def _influence_after_edge_loop_update(self, infl_list, panel_name, new_edge_ids):
        """
            Update the list of parameter\constraint influence with the new edge ids of given panel.
            
            flipped -- indicates if in the new edges start & end vertices have been swapped
        """
        for infl_id in range(len(infl_list)):
            if infl_list[infl_id]['panel'] == panel_name:
                # update
                edge_list = infl_list[infl_id]['edge_list']
                for edge_list_id in range(len(edge_list)):
                    if isinstance(edge_list[edge_list_id], int):  # Simple edge id lists in curvature params
                        old_id = edge_list[edge_list_id]
                        edge_list[edge_list_id] = new_edge_ids[old_id]
                    elif isinstance(edge_list[edge_list_id]['id'], list):  # Meta-edge in length parameters  & constraints
                        for i in range(len(edge_list[edge_list_id]['id'])):
                            old_id = edge_list[edge_list_id]['id'][i]
                            edge_list[edge_list_id]['id'][i] = new_edge_ids[old_id]
                    else:  # edge description in length parameters & constraints
                        old_id = edge_list[edge_list_id]['id']
                        edge_list[edge_list_id]['id'] = new_edge_ids[old_id]
                
    def _update_pattern_by_param_values(self):
        """
        Recalculates vertex positions and edge curves according to current
        parameter values
        (!) Assumes that the current pattern is a template:
                with all the parameters equal to defaults!
        """
        for parameter in self.spec['parameter_order']:
            value = self.parameters[parameter]['value']
            param_type = self.parameters[parameter]['type']
            if param_type not in self.parameter_defaults:
                raise ValueError("Incorrect parameter type. Alowed are "
                                 + self.parameter_defaults.keys())

            for panel_influence in self.parameters[parameter]['influence']:
                for edge in panel_influence['edge_list']:
                    if param_type == 'length':
                        self._extend_edge(panel_influence['panel'], edge, value)
                    elif param_type == 'additive_length':
                        self._extend_edge(panel_influence['panel'], edge, value, multiplicative=False)
                    elif param_type == 'curve':
                        self._curve_edge(panel_influence['panel'], edge, value)
        # finally, ensure secified constraints are held
        self._apply_constraints()    

    def _restore_template(self, params_to_default=True):
        """Restore pattern to it's state with all parameters having default values
            Recalculate vertex positions, edge curvatures & snap values to 1
        """
        # Follow process backwards
        self._invert_constraints()

        for parameter in reversed(self.spec['parameter_order']):
            value = self.parameters[parameter]['value']
            param_type = self.parameters[parameter]['type']
            if param_type not in self.parameter_defaults:
                raise ValueError("Incorrect parameter type. Alowed are "
                                 + self.parameter_defaults.keys())

            for panel_influence in reversed(self.parameters[parameter]['influence']):
                for edge in reversed(panel_influence['edge_list']):
                    if param_type == 'length':
                        self._extend_edge(panel_influence['panel'], edge, self._invert_value(value))
                    elif param_type == 'additive_length':
                        self._extend_edge(panel_influence['panel'], edge, 
                                          self._invert_value(value, multiplicative=False), 
                                          multiplicative=False)
                    elif param_type == 'curve':
                        self._curve_edge(panel_influence['panel'], edge, self._invert_value(value))
            
            # restore defaults
            if params_to_default:
                if isinstance(value, list):
                    self.parameters[parameter]['value'] = [self.parameter_defaults[param_type] for _ in value]
                else:
                    self.parameters[parameter]['value'] = self.parameter_defaults[param_type]

    def _extend_edge(self, panel_name, edge_influence, value, multiplicative=True):
        """
            Shrinks/elongates a given edge or edge collection of a given panel. Applies equally
            to straight and curvy edges tnks to relative coordinates of curve controls
            Expects
                * each influenced edge to supply the elongatoin direction
                * scalar scaling_factor
            'multiplicative' parameter controls the type of extention:
                * if True, value is treated as a scaling factor of the edge or edge projection -- default
                * if False, value is added to the edge or edge projection
        """
        if isinstance(value, list):
            raise ValueError("Multiple scaling factors are not supported")

        verts_ids, verts_coords, target_line, _ = self._meta_edge(panel_name, edge_influence)

        # calc extention pivot
        if edge_influence['direction'] == 'end':
            fixed = verts_coords[0]  # start is fixed
        elif edge_influence['direction'] == 'start':
            fixed = verts_coords[-1]  # end is fixed
        elif edge_influence['direction'] == 'both':
            fixed = (verts_coords[0] + verts_coords[-1]) / 2
        else:
            raise RuntimeError('Unknown edge extention direction {}'.format(edge_influence['direction']))

        # move verts 
        # * along target line that sits on fixed point (correct sign & distance along the line)
        verts_projection = np.empty(verts_coords.shape)
        for i in range(verts_coords.shape[0]):
            verts_projection[i] = (verts_coords[i] - fixed).dot(target_line) * target_line

        if multiplicative:
            # * to match the scaled projection (correct point of application -- initial vertex position)
            new_verts = verts_coords - (1 - value) * verts_projection
        else:
            # * to match the added projection: 
            # still need projection to make sure the extention derection is corect relative to fixed point
            
            # normalize first
            for i in range(verts_coords.shape[0]):
                norm = np.linalg.norm(verts_projection[i])
                if not np.isclose(norm, 0):
                    verts_projection[i] /= norm

            # zero projections were not normalized -- they will zero-out the effect
            new_verts = verts_coords + value * verts_projection

        # update in the initial structure
        panel = self.pattern['panels'][panel_name]
        for ni, idx in enumerate(verts_ids):
            panel['vertices'][idx] = new_verts[ni].tolist()

    def _curve_edge(self, panel_name, edge, scaling_factor):
        """
            Updated the curvature of an edge accoding to scaling_factor.
            Can only be applied to edges with curvature information
            scaling_factor can be
                * scalar -- only the Y of control point is changed
                * 2-value list -- both coordinated of control are updated
        """
        panel = self.pattern['panels'][panel_name]
        if 'curvature' not in panel['edges'][edge]:
            raise ValueError('Applying curvature scaling to non-curvy edge '
                             + str(edge) + ' of ' + panel_name)
        control = panel['edges'][edge]['curvature']

        if isinstance(scaling_factor, list):
            control = [
                control[0] * scaling_factor[0],
                control[1] * scaling_factor[1]
            ]
        else:
            control[1] *= scaling_factor

        panel['edges'][edge]['curvature'] = control

    def _invert_value(self, value, multiplicative=True):
        """If value is a list, return a list with each value inverted.
            'multiplicative' parameter controls the type of inversion:
                * if True, returns multiplicative inverse (1/value) == default
                * if False, returns additive inverse (-value)
        """
        if multiplicative:
            if isinstance(value, list):
                if any(np.isclose(value, 0)):
                    raise ZeroDivisionError('Zero value encountered while restoring multiplicative parameter.')
                return map(lambda x: 1 / x, value)
            else:
                if np.isclose(value, 0):
                    raise ZeroDivisionError('Zero value encountered while restoring multiplicative parameter.')
                return 1 / value
        else:
            if isinstance(value, list):
                return map(lambda x: -x, value)
            else:
                return -value

    def _apply_constraints(self):
        """Change the pattern to adhere to constraints if given in pattern spec
            Assumes no zero-length edges exist"""
        if 'constraints' not in self.spec:
            return 

        for constraint_n in self.spec['constraints']:  # order preserved as it's a list
            constraint = self.spec['constraints'][constraint_n]
            constraint_type = constraint['type']
            if constraint_type not in self.constraint_types:
                raise ValueError("Incorrect constraint type. Alowed are "
                                 + self.constraint_types)

            if constraint_type == 'length_equality':
                # get all length of the affected (meta) edges
                target_len = []
                for panel_influence in constraint['influence']:
                    for edge in panel_influence['edge_list']:
                        _, _, _, length = self._meta_edge(panel_influence['panel'], edge)
                        edge['length'] = length
                        target_len.append(length)
                if len(target_len) == 0:
                    return
                # target as mean of provided edges
                target_len = sum(target_len) / len(target_len)  

                # calculate scaling factor for every edge to match max length
                # & update edges with it
                for panel_influence in constraint['influence']:
                    for edge in panel_influence['edge_list']:
                        scaling = target_len / edge['length'] 
                        if not np.isclose(scaling, 1):
                            edge['value'] = scaling
                            self._extend_edge(panel_influence['panel'], edge, edge['value'])

    def _invert_constraints(self):
        """Restore pattern to the state before constraint was applied"""
        if 'constraints' not in self.spec:
            return 

        # follow the process backwards
        for constraint_n in reversed(self.spec['constraint_order']):  # order preserved as it's a list
            constraint = self.spec['constraints'][constraint_n]
            constraint_type = constraint['type']
            if constraint_type not in self.constraint_types:
                raise ValueError("Incorrect constraint type. Alowed are "
                                 + self.constraint_types)

            if constraint_type == 'length_equality':
                # update edges with invertes scaling factor
                for panel_influence in constraint['influence']:
                    for edge in panel_influence['edge_list']:
                        scaling = self._invert_value(edge['value'])
                        self._extend_edge(panel_influence['panel'], edge, scaling)
                        edge['value'] = 1

    def _meta_edge(self, panel_name, edge_influence):
        """Returns info for the given edge or meta-edge in inified form"""

        panel = self.pattern['panels'][panel_name]
        edge_ids = edge_influence['id']
        if isinstance(edge_ids, list):
            # meta-edge
            # get all vertices in order
            verts_ids = [panel['edges'][edge_ids[0]]['endpoints'][0]]  # start
            for edge_id in edge_ids:
                verts_ids.append(panel['edges'][edge_id]['endpoints'][1])  # end vertices
        else:
            # single edge
            verts_ids = panel['edges'][edge_ids]['endpoints']

        verts_coords = []
        for idx in verts_ids:
            verts_coords.append(panel['vertices'][idx])
        verts_coords = np.array(verts_coords)

        # extention line
        if 'along' in edge_influence:
            target_line = edge_influence['along']
        else:
            target_line = verts_coords[-1] - verts_coords[0] 
        target_line = np.array(target_line, dtype=float)  # https://stackoverflow.com/questions/50625975/typeerror-ufunc-true-divide-output-typecode-d-could-not-be-coerced-to-pro

        if np.isclose(np.linalg.norm(target_line), 0):
            raise ZeroDivisionError('target line is zero ' + str(target_line))
        else:
            target_line /= np.linalg.norm(target_line)

        return verts_ids, verts_coords, target_line, target_line.dot(verts_coords[-1] - verts_coords[0])

    def _invalidate_all_values(self):
        """Sets all values of params & constraints to None if not set already
            Useful in direct updates of pattern panels"""

        updated_once = False
        for parameter in self.parameters:
            if self.parameters[parameter]['value'] is not None:
                self.parameters[parameter]['value'] = None
                updated_once = True
        
        if 'constraints' in self.spec:
            for constraint in self.spec['constraints']:
                for edge_collection in self.spec['constraints'][constraint]['influence']:
                    for edge in edge_collection['edge_list']:
                        if edge['value'] is not None: 
                            edge['value'] = None
                            updated_once = True
        if updated_once:
            # only display worning if some new invalidation happened
            print('ParametrizedPattern::Warning::Parameter (& constraints) values are invalidated')

    # ---------- Randomization -------------
    def _randomize_pattern(self):
        """Robustly randomize current pattern"""
        # restore template state before making any changes to parameters
        self._restore_template(params_to_default=False)

        spec_backup = copy.deepcopy(self.spec)
        self._randomize_parameters()
        self._update_pattern_by_param_values()
        for _ in range(100):  # upper bound on trials to avoid infinite loop
            if not self.is_self_intersecting():
                break

            print('Warning::Randomized pattern is self-intersecting. Re-try..')
            self._restore(spec_backup)
            # Try again
            self._randomize_parameters()
            self._update_pattern_by_param_values()

    def _new_value(self, param_range):
        """Random value within range given as an iteratable"""
        value = random.uniform(param_range[0], param_range[1])
        # prevent non-reversible zero values
        if abs(value) < 1e-2:
            value = 1e-2 * (-1 if value < 0 else 1)
        return value

    def _randomize_parameters(self):
        """
        Sets new random values for the pattern parameters
        Parameter type agnostic
        """
        for parameter in self.parameters:
            param_ranges = self.parameters[parameter]['range']

            # check if parameter has multiple values (=> multiple ranges) like for curves
            if isinstance(self.parameters[parameter]['value'], list): 
                values = []
                for param_range in param_ranges:
                    values.append(self._new_value(param_range))
                self.parameters[parameter]['value'] = values
            else:  # simple 1-value parameter
                self.parameters[parameter]['value'] = self._new_value(param_ranges)


# ---------- test -------------
if __name__ == "__main__":
    import customconfig
    from pattern.wrappers import VisPattern

    # np.set_printoptions(precision=4, suppress=True)

    system_config = customconfig.Properties('./system.json')
    base_path = system_config['output']
    pattern = ParametrizedPattern(os.path.join(system_config['templates_path'], 'basic tee', 'tee_rotated.json'))
    # pattern = ParametrizedPattern(os.path.join(system_config['templates_path'], 'skirts', 'skirt_4_panels.json'))
    # pattern = BasicPattern(os.path.join(system_config['datasets_path'], 'data_1000_tee_200527-14-50-42_regen_200612-16-56-43', 'tee_8O9CU32Q8G', 'specification.json'))
    # pattern_init = BasicPattern(os.path.join(base_path, 'nn_pred_data_1000_tee_200527-14-50-42_regen_200612-16-56-43201106-14-46-31', 'test', 'tee_8O9CU32Q8G', 'specification.json'))
    # pattern_predicted = BasicPattern(os.path.join(base_path, 'nn_pred_data_1000_tee_200527-14-50-42_regen_200612-16-56-43201106-14-46-31', 'test', 'tee_8O9CU32Q8G', '_predicted_specification.json'))
    # pattern = VisPattern()
    empty_pattern = BasicPattern()
    print(pattern.panel_order())

    # print(pattern.stitches_as_tags())

    # print(len(pattern.pattern_as_tensors(with_placement=True, with_stitches=True, with_stitch_tags=True)))

    # tensor, num_panels, rot, transl, stitches, stitch_num, stitch_tags = pattern.pattern_as_tensors(with_placement=True, with_stitches=True, with_stitch_tags=True)

    # empty_pattern.pattern_from_tensors(tensor, rot, transl, stitches, padded=True)
    # print(pattern.pattern['stitches'])
    # print(empty_pattern.panel_order())

    pattern.name = pattern.name + '_save_pattern_order' + '_' + datetime.now().strftime('%y%m%d-%H-%M-%S')
    # empty_pattern.serialize(system_config['output'], to_subfolder=True)
    pattern.serialize(system_config['output'], to_subfolder=True)





import numpy as np
from scipy.spatial.transform import Rotation as R

# Custom
from pattern.core import BasicPattern
from pattern.wrappers import VisPattern
from .base import BaseComponent
from .interface import Interface

class Component(BaseComponent):
    """Garment element (or whole piece) composed of simpler connected garment elements"""

    def __init__(self, name) -> None:
        super().__init__(name)

        self.subs = []  # list of generative sub-components

    # Operations -- update object in-place
    # All return self object to allow chained operations

    # Placements
    def pivot_3D(self):
        """Pivot of a component as a block

            NOTE: The relation of pivots of subblocks needs to be 
            preserved in any placement operations on components
        """
        mins, maxes = self.bbox3D()

        return np.array(((mins[0] + maxes[0]) / 2, maxes[1], (mins[-1] + maxes[-1]) / 2))

    def translate_by(self, delta_vector):
        """Translate component by a vector"""
        for subs in self._get_subcomponents():
            subs.translate_by(delta_vector)
        return self
    
    def translate_to(self, new_translation):
        """Set panel translation to be exactly that vector"""
        pivot = self.pivot_3D()
        for subs in self._get_subcomponents():
            sub_pivot = subs.pivot_3D()
            subs.translate_to(np.asarray(new_translation) + (sub_pivot - pivot))
        return self

    def rotate_by(self, delta_rotation:R):
        """Rotate component by a given rotation"""
        pivot = self.pivot_3D()
        for subs in self._get_subcomponents():
            # With preserving relationships between components
            rel = subs.pivot_3D() - pivot
            rel_rotated = delta_rotation.apply(rel) 
            subs.rotate_by(delta_rotation)
            subs.translate_by(rel_rotated - rel)
        return self
    
    def rotate_to(self, new_rot):
        raise NotImplementedError(
            f'Component::Error::rotate_to is not supported on component level.'
            'Use relative <rotate_by()> method instead')

    def place_below(self, comp: BaseComponent, gap=2):
        """Place below the provided component"""
        other_bbox = comp.bbox3D()
        curr_bbox = self.bbox3D()

        self.translate_by([0, other_bbox[0][1] - curr_bbox[1][1] - gap, 0])

        return self

    def place_by_interface(self, 
                            self_interface:Interface, 
                            out_interface:Interface, 
                            gap=2):
        """Adjust the placement of component acconding to the connectivity instuction        
        """
        
        # Alight translation
        self_bbox = self_interface.bbox_3d()
        out_bbox = out_interface.bbox_3d()
        mid_out = (out_bbox[1] + out_bbox[0]) / 2
        mid_self = (self_bbox[1] + self_bbox[0]) / 2

        # Add a gap outside of the current 
        full_bbox = self.bbox3D()
        center = (full_bbox[0] + full_bbox[1]) / 2
        gap_dir = mid_self - center
        gap_dir = gap * gap_dir / np.linalg.norm(gap_dir)
        
        diff = mid_out - (mid_self + gap_dir)
        
        self.translate_by(diff)

        # NOTE: Norm evaluation of vertex set will fail 
        # for the alignment of 2D panels, where they are likely
        # to be in one line or in a panel plane instead of 
        # the interface place -- so I'm not using norms for gap esitmation

        return self

    # Mirror
    def mirror(self, axis=[0, 1]):
        """Swap this component with it's mirror image by recursively mirroring sub-components
        
            Axis specifies 2D axis to swap around: Y axis by default
        """
        for subs in self._get_subcomponents():
            subs.mirror(axis)
        return self

    # Build the component -- get serializable representation
    def assembly(self):
        """Construction process of the garment component
        
        Returns: simulator friendly description of component sewing pattern
        """
        spattern = VisPattern()
        spattern.name = self.name

        subs = self._get_subcomponents()
        if not subs:
            return spattern

        # Simple merge of sub-component representations
        for sub in subs:
            sub_raw = sub().pattern

            # simple merge of panels
            spattern.pattern['panels'] = {**spattern.pattern['panels'], **sub_raw['panels']}

            # of stitches
            spattern.pattern['stitches'] += sub_raw['stitches']

        spattern.pattern['stitches'] += self.stitching_rules.assembly()

        return spattern   

    # Utilities
    def bbox3D(self):
        """Evaluate 3D bounding box of the current component"""
        
        subs = self._get_subcomponents()
        bboxes = [s.bbox3D() for s in subs]

        mins = np.vstack([b[0] for b in bboxes])
        maxes = np.vstack([b[1] for b in bboxes])

        return mins.min(axis=0), maxes.max(axis=0)

    # Subcomponents
    def _get_subcomponents(self):
        """Unique set of subcomponents defined in the self.subs list or as attributes of the object"""

        all_attrs = [getattr(self, name) for name in dir(self) if name[:2] != '__' and name[-2:] != '__']
        return list(set([att for att in all_attrs if isinstance(att, BaseComponent)] + self.subs))

